"""Edge-case coverage for DataTable internals and error branches."""

from collections import namedtuple
import logging
import sys
import types
from pathlib import Path

import pytest

import soigia.datatable as dtpkg
from soigia.datatable import DataTable


def test_values_result_branches():
    records = dtpkg._ValuesResult(
        [
            {"name": "alice", "age": 19},
            {"name": "bob", "age": 24},
            {"name": "alice", "age": 19},
            ("x", "y"),
            ["x", "y"],
            123,
        ]
    )

    assert records.filter(lambda row: isinstance(row, dict) and row.get("name") == "alice") == [
        {"name": "alice", "age": 19},
        {"name": "alice", "age": 19},
    ]
    assert records.exclude(lambda row: isinstance(row, dict) and row.get("name") == "alice") == [
        {"name": "bob", "age": 24},
        ("x", "y"),
        ["x", "y"],
        123,
    ]
    assert records.order_by(lambda row: str(row)) == _sorted_records(records, key=lambda row: str(row))
    assert dtpkg._ValuesResult([1, 1, 2, 3, 2]).distinct() == [1, 2, 3]
    assert dtpkg._ValuesResult([{"a": 1}, {"a": 1}, {"a": 2}]).distinct("a") == [{"a": 1}, {"a": 2}]

    with pytest.raises(TypeError):
        records.filter(predicate="bad")
    with pytest.raises(TypeError):
        records.exclude(predicate="bad")
    with pytest.raises(TypeError):
        records.order_by(123)


def _sorted_records(records, key):
    return dtpkg._ValuesResult(sorted(records, key=key))


def test_no_logs_context_branches():
    ctx = dtpkg._NoLogsContext("soigia.edge", level=logging.INFO)
    with ctx:
        pass

    ctx2 = dtpkg._NoLogsContext("soigia.edge", level=logging.INFO)
    with pytest.raises(AssertionError):
        with ctx2:
            logging.getLogger("soigia.edge").warning("hello")


def test_mini_faker_and_resolve_fake_value_branches():
    fake = dtpkg._MiniFaker(seed=123)
    fake.seed_instance(123)
    assert fake.name()
    assert fake.first_name()
    assert fake.last_name()
    assert fake.city()
    assert fake.company()
    assert fake.word()
    assert fake.words(2)
    assert fake.sentence(4).endswith(".")
    assert fake.text(20)
    assert "@" in fake.email()
    assert fake.phone_number().startswith("+84-")
    assert isinstance(fake.uuid4(), str)
    assert isinstance(fake.random_int(1, 3), int)
    assert isinstance(fake.pyfloat(), float)
    assert fake.date().tzinfo is not None
    assert fake.datetime().tzinfo is not None

    assert dtpkg._resolve_fake_value(fake, lambda gen: gen.name()) != ""
    assert dtpkg._resolve_fake_value(fake, ("name", {}))
    assert dtpkg._resolve_fake_value(fake, {"provider": "city"}) in fake._cities
    assert dtpkg._resolve_fake_value(fake, "name")
    assert dtpkg._resolve_fake_value(fake, 123) == 123


def test_factory_conversion_branches():
    Point = namedtuple("Point", "x y")

    class Dictable:
        def dict(self):
            return {"name": "dictable"}

    class AttrObject:
        def __init__(self):
            self.name = "attr"
            self.age = 99
            self._private = "hidden"

    assert dtpkg._factory_item_to_record({"name": "alice"}) == {"name": "alice"}
    assert dtpkg._factory_item_to_record(Point(1, 2)) == {"x": 1, "y": 2}
    assert dtpkg._factory_item_to_record(Dictable()) == {"name": "dictable"}
    assert dtpkg._factory_item_to_record(AttrObject()) == {"name": "attr", "age": 99}

    with pytest.raises(TypeError):
        dtpkg._factory_item_to_record(123)

    class BatchFactory:
        @classmethod
        def build_batch(cls, size=1, **kwargs):
            return [{"name": f"row_{i}", "kind": kwargs.get("kind", "a")} for i in range(size)]

    assert dtpkg._build_factory_rows(BatchFactory, 2, kind="b") == [
        {"name": "row_0", "kind": "b"},
        {"name": "row_1", "kind": "b"},
    ]
    assert dtpkg._build_factory_rows(lambda **kwargs: {"name": kwargs.get("name", "x")}, 2, name="y") == [
        {"name": "y"},
        {"name": "y"},
    ]
    with pytest.raises(TypeError):
        dtpkg._build_factory_rows(object(), 1)


def test_database_connector_branches(monkeypatch):
    connector = dtpkg.DatabaseConnector(connection_string="sqlite:///example.db", mysql="mysql:///mysql.db")
    assert connector._engine_url("mysql") == "mysql:///mysql.db"
    assert connector._engine_url("postgres", connection_string="postgresql:///db") == "postgresql:///db"

    monkeypatch.setenv("SOIGIA_POSTGRES_URL", "postgresql:///envdb")
    connector2 = dtpkg.DatabaseConnector()
    assert connector2._engine_url("postgres") == "postgresql:///envdb"

    with pytest.raises(ValueError):
        dtpkg.DatabaseConnector()._engine_url("oracle")

    monkeypatch.setattr(dtpkg, "create_engine", None)
    with pytest.raises(ImportError):
        connector2._engine("sqlite")


def test_join_and_schema_error_branches():
    left = DataTable({"id": [1, 2], "name": ["alice", "bob"], "age": ["19", "24"]})
    right = DataTable({"id": [1, 2], "city": ["hn", "sg"]})

    assert left.join(right, on="id").shape[0] == 2
    assert left.join(right, left_on="id", right_on="id").shape[0] == 2
    assert left.join(DataTable({"id": [1], "city": ["hn"]})).shape[0] == 1

    with pytest.raises(ValueError):
        left.join(right, on=[])
    with pytest.raises(ValueError):
        left.join(right, left_on="id")
    with pytest.raises(ValueError):
        left.join(right, left_on=["id"], right_on=["id", "extra"])
    with pytest.raises(ValueError):
        DataTable({"a": [1]}).join(DataTable({"b": [1]}))

    with pytest.raises(ValueError):
        left.validate_schema({"extra": {"required": True}}, strict=True)
    with pytest.raises(KeyError):
        left.validate_schema({"missing": {"required": True}})
    with pytest.raises(ValueError):
        DataTable({"name": ["alice", None]}).validate_schema({"name": {"nullable": False}})
    with pytest.raises(TypeError):
        left.validate_schema({"age": {"type": int}})
    numeric = DataTable({"age": [19, 24]})
    with pytest.raises(ValueError):
        left.validate_schema({"name": {"choices": ["carol"]}})
    with pytest.raises(ValueError):
        numeric.validate_schema({"age": {"min": 20}})
    with pytest.raises(ValueError):
        numeric.validate_schema({"age": {"max": 20}})
    with pytest.raises(ValueError):
        left.validate_schema({"name": {"regex": "^z"}})
    with pytest.raises(ValueError):
        left.validate_schema({"name": {"length": 2}})


def test_type_coercion_render_and_snapshot_branches():
    table = DataTable({"age": ["19", "24"], "created_at": ["2026-01-01", "2026-01-02"]})
    coerced = table.coerce_types({"age": "int64", "created_at": "datetime"})
    assert coerced["age"].dtype.kind in {"i", "u"}
    assert str(coerced["created_at"].dtype).startswith("datetime64")

    coerced_date = table.coerce_types({"created_at": "date"})
    assert coerced_date["created_at"].iloc[0].isoformat() == "2026-01-01"

    with pytest.raises(KeyError):
        table.coerce_types({"missing": "int64"})

    with pytest.raises(ValueError):
        table.render_table("excel")

    with pytest.raises(ValueError):
        table.render_table("unknown")

    table.init_versions("initial")
    snapshot = table.snapshot("snap")
    assert snapshot.version_count() == 2
    assert snapshot.current_version()["label"] == "snap"

    auto = table.auto_version("auto")
    assert auto.is_auto_versioning() is True
    assert auto.disable_auto_version().is_auto_versioning() is False


def test_manager_delegation_branches():
    table = DataTable({"id": [1, 2, 3], "name": ["alice", "bob", "carol"], "age": [19, 24, 31]})
    manager = table.objects

    assert manager.all().count() == 3
    assert manager.none().count() == 0
    assert manager.values_list("name", flat=True).first() == "alice"
    assert manager.select("name").to_df().columns.tolist() == ["name"]
    assert manager.exclude_fields("age").to_df().columns.tolist() == ["id", "name"]
    assert manager.get(name="bob")["age"].iloc[0] == 24
    assert manager.first()["name"].iloc[0] == "alice"
    assert manager.last()["name"].iloc[0] == "carol"
    assert manager.earliest("age")["name"].iloc[0] == "alice"
    assert manager.latest("age")["name"].iloc[0] == "carol"
    assert manager.paginate(limit=2).count() == 2
    assert manager.group_by("name").count().count() == 3
    assert manager.aggregate(total=("age", "sum")) == {"total": 74}
    assert manager.sum("age") == {"age": 74}
    assert manager.avg("age") == {"age": 24.666666666666668}
    assert manager.min("age") == {"age": 19}
    assert manager.max("age") == {"age": 31}
    assert manager.values("name").first() == {"name": "alice"}
    assert manager.diff_columns_against(table)["column_order_changed"] is False
    assert manager.diff_added_rows(table).shape[0] == 0
    assert manager.diff_removed_rows(table).shape[0] == 0
    assert manager.diff_unchanged_rows(table).shape[0] == 3
    assert manager.diff_changed_rows(table).shape[0] == 0
    assert manager.diff_changed_cells(table).shape[0] == 0
    assert manager.diff_report(table)["has_changes"] is False

    created = manager.create(name="dave", age=40)
    assert created.to_rows() == [{"name": "dave", "age": 40}]
    assert manager.append_row(name="dave", age=40).shape[0] == 4
    assert manager.bulk_create([{"name": "x", "age": 1}]).shape[0] == 4
    assert manager.update(age=25) == 3
    assert manager.delete() == 3
    updated_df, count = manager.bulk_update({"age": 25}, return_count=True, age__gte=20)
    assert count == 2
    assert updated_df.shape[0] == 3
    assert manager.save().equals(table)
    assert manager.get_or_create(name="bob", defaults={"age": 24})[1] is False
    assert manager.update_or_create(name="new", defaults={"age": 18})[1] is True

    joined = manager.join(DataTable({"id": [1, 2, 3], "city": ["hn", "sg", "dn"]}), on="id").to_df()
    assert "city" in joined.columns


def test_fake_faker_and_csv_from_helpers(monkeypatch):
    fake_module = types.ModuleType("faker")

    class FakeFaker:
        def __init__(self, locale):
            self.locale = locale
            self.seeded = None

        def seed_instance(self, seed):
            self.seeded = seed

        def name(self):
            return "fake name"

        def city(self):
            return "fake city"

    fake_module.Faker = FakeFaker
    monkeypatch.setitem(sys.modules, "faker", fake_module)

    fake = dtpkg._build_faker(locale="vi_VN", seed=7)
    assert isinstance(fake, FakeFaker)
    assert fake.locale == "vi_VN"
    assert fake.seeded == 7
    assert dtpkg._resolve_fake_value(fake, ("name", {})) == "fake name"
    assert dtpkg._resolve_fake_value(fake, {"provider": "city"}) == "fake city"
    assert dtpkg._resolve_fake_value(fake, "name") == "fake name"
    assert dtpkg._resolve_fake_value(fake, 123) == 123

    values = dtpkg._ValuesResult([("x", "y"), ["x", "y"], 123, {"a": 1}, {"a": 1}])
    assert values.distinct() == [("x", "y"), 123, {"a": 1}]
    assert dtpkg._ValuesResult([{"a": 1}, {"a": 1}, {"a": 2}]).distinct("a") == [{"a": 1}, {"a": 2}]

    csv_path = Path("tests") / "_datatable_coverage.csv"
    csv_path.write_text("a,b\n1,2\n", encoding="utf-8")
    try:
        loaded = DataTable.from_csv(csv_path)
        assert loaded.to_dict(orient="records") == [{"a": 1, "b": 2}]
        assert DataTable.from_dict({"x": [1, 2]}).shape == (2, 1)
    finally:
        if csv_path.exists():
            csv_path.unlink()


def test_database_helpers_and_classmethod_wrappers(monkeypatch):
    calls = []

    def fake_query(self, sql, backend, connection_string=None, **kwargs):
        calls.append((backend, sql, connection_string, kwargs))
        return DataTable({"backend": [backend], "sql": [sql]})

    monkeypatch.setattr(dtpkg.DatabaseConnector, "query", fake_query)

    connector = dtpkg.DatabaseConnector(
        connection_string="sqlite:///base.db",
        mysql="mysql:///db",
        maria="maria:///db",
        oracle="oracle:///db",
        postgres="postgres:///db",
        sqlite="sqlite:///db",
    )
    assert connector.from_sql("select 1").iloc[0]["backend"] == "sqlite"
    assert connector.from_mysql_db("select 1").iloc[0]["backend"] == "mysql"
    assert connector.from_maria_db("select 1").iloc[0]["backend"] == "maria"
    assert connector.from_oracle_db("select 1").iloc[0]["backend"] == "oracle"
    assert connector.from_postgres_db("select 1").iloc[0]["backend"] == "postgres"
    assert connector.from_sqlite_db("select 1").iloc[0]["backend"] == "sqlite"

    default = dtpkg.db_connector(connection_string="sqlite:///default.db")
    assert isinstance(default, dtpkg.DatabaseConnector)
    monkeypatch.setattr(dtpkg, "_DEFAULT_CONNECTOR", connector)

    assert dtpkg.from_sql("select 1").iloc[0]["backend"] == "sqlite"
    assert dtpkg.from_mysql_db("select 1").iloc[0]["backend"] == "mysql"
    assert dtpkg.from_maria_db("select 1").iloc[0]["backend"] == "maria"
    assert dtpkg.from_oracle_db("select 1").iloc[0]["backend"] == "oracle"
    assert dtpkg.from_postgres_db("select 1").iloc[0]["backend"] == "postgres"
    assert dtpkg.from_sqlite_db("select 1").iloc[0]["backend"] == "sqlite"

    monkeypatch.setattr(dtpkg, "from_sql", lambda sql, backend="sqlite", connection_string=None, **kwargs: DataTable({"kind": [backend]}))
    monkeypatch.setattr(dtpkg, "from_mysql_db", lambda sql, connection_string=None, **kwargs: DataTable({"kind": ["mysql"]}))
    monkeypatch.setattr(dtpkg, "from_maria_db", lambda sql, connection_string=None, **kwargs: DataTable({"kind": ["maria"]}))
    monkeypatch.setattr(dtpkg, "from_oracle_db", lambda sql, connection_string=None, **kwargs: DataTable({"kind": ["oracle"]}))
    monkeypatch.setattr(dtpkg, "from_postgres_db", lambda sql, connection_string=None, **kwargs: DataTable({"kind": ["postgres"]}))
    monkeypatch.setattr(dtpkg, "from_sqlite_db", lambda sql, connection_string=None, **kwargs: DataTable({"kind": ["sqlite"]}))

    assert DataTable.from_sql("select 1")["kind"].iloc[0] == "sqlite"
    assert DataTable.from_mysql_db("select 1")["kind"].iloc[0] == "mysql"
    assert DataTable.from_maria_db("select 1")["kind"].iloc[0] == "maria"
    assert DataTable.from_oracle_db("select 1")["kind"].iloc[0] == "oracle"
    assert DataTable.from_postgres_db("select 1")["kind"].iloc[0] == "postgres"
    assert DataTable.from_sqlite_db("select 1")["kind"].iloc[0] == "sqlite"

    assert len(calls) >= 6


def test_deep_diff_and_version_branches(monkeypatch):
    class WeirdEq:
        def __eq__(self, other):
            return Boolish()

    class Boolish:
        def __bool__(self):
            return True

    class EqRaises:
        def __eq__(self, other):
            raise RuntimeError("boom")

        def __repr__(self):
            return "<eq-raises>"

    monkeypatch.setattr(dtpkg.pd, "isna", lambda value: False)
    assert dtpkg.DataTable._cell_equal(WeirdEq(), WeirdEq()) is True
    assert dtpkg.DataTable._cell_equal(EqRaises(), EqRaises()) is True

    class NaRaises:
        pass

    monkeypatch.setattr(dtpkg.pd, "isna", lambda value: (_ for _ in ()).throw(RuntimeError("na")))
    assert dtpkg.DataTable._cell_equal(NaRaises(), NaRaises()) is False

    base = DataTable({"id": [1, 2], "name": ["alice", "bob"], "age": ["19", "24"]})
    base.init_versions("start")
    assert "records" not in base.version_history()[0]
    assert base.version_history(include_data=True)[0]["records"] == [{"id": 1, "name": "alice", "age": "19"}, {"id": 2, "name": "bob", "age": "24"}]

    auto_base = DataTable({"id": [1, 2], "name": ["alice", "bob"], "age": ["19", "24"]}).auto_version("auto")
    grown = auto_base.join(DataTable({"id": [1, 2, 3], "city": ["hn", "sg", "dn"]}), on="id", how="outer")
    shrunk = auto_base.join(DataTable({"id": [1], "city": ["hn"]}), on="id", how="inner")

    assert auto_base.is_auto_versioning() is True
    assert auto_base.disable_auto_version().is_auto_versioning() is False
    assert auto_base.version_count() == 1
    assert grown.version_count() == 2
    assert grown.version_history()[0]["label"] == "initial"
    assert grown.current_version()["label"] == "auto"
    assert grown.get_version(0).shape == (2, 3)
    assert grown.get_version(-1).shape == (3, 4)
    assert grown.restore_version(0).equals(grown.get_version(0))

    assert grown.diff_versions() ["row_delta"] == 1
    assert grown.diff_records()["row_delta"] == 1
    assert grown.diff_columns()["column_order_changed"] is True
    assert grown.diff_summary(key_columns=["id"])["changed_rows"] >= 0

    keyed_cells = grown.diff_cells(key_columns=["id"])
    assert isinstance(keyed_cells, dict)
    assert grown.diff_cells(old=0, new=-1, key_columns=["id"])["added_rows"]
    assert shrunk.diff_cells(old=0, new=-1)["removed_rows"]
    with pytest.raises(KeyError):
        grown.diff_cells(key_columns=["missing"])

    report = grown.diff_report(shrunk, key_columns=["id"])
    assert report["has_changes"] is True
    assert "column_order_changed" in report["column_diff"]
    assert report["same_rows"].shape[0] >= 0
    assert report["changed_cells"].shape[0] >= 0

    comparison = grown.compare_frames(shrunk, key_columns=["id"])
    assert comparison["same_count"] >= 0
    assert grown.same_rows(shrunk, key_columns=["id"]).shape[0] >= 0
    diff_rows = grown.different_rows(shrunk, key_columns=["id"])
    assert diff_rows["changed_count"] >= 0
    assert grown.diff_rows(shrunk, key_columns=["id"])["left_only_count"] >= 0
    assert "shared_columns" in grown.diff_columns_against(shrunk)
    assert grown.diff_added_rows(shrunk, key_columns=["id"]).shape[0] >= 0
    assert grown.diff_removed_rows(shrunk, key_columns=["id"]).shape[0] >= 0
    assert grown.diff_unchanged_rows(shrunk, key_columns=["id"]).shape[0] >= 0
    assert grown.diff_changed_rows(shrunk, key_columns=["id"]).shape[0] >= 0
    assert grown.diff_changed_cells(shrunk, key_columns=["id"]).shape[0] >= 0
    assert grown.diff_changed_cells(shrunk).shape[0] >= 0
    assert grown.different_rows(shrunk)["changed_count"] >= 0

    with pytest.raises(KeyError):
        grown.compare_frames(shrunk, key_columns=["missing"])

    with pytest.raises(IndexError):
        grown.diff_cells(old=999, new=-1)
    with pytest.raises(IndexError):
        grown.get_version(999)
    with pytest.raises(KeyError):
        grown.get_version("missing")
    assert grown.diff_summary()["added_rows"] >= 0


def test_queryset_and_grouped_branches(monkeypatch):
    table = DataTable(
        {
            "id": [1, 2, 3],
            "name": ["alice", "bob", "carol"],
            "age": [19, 24, 31],
            "active": [True, False, True],
        }
    )
    qs = dtpkg._ORMQuerySet(table)

    assert isinstance(qs["name"], dtpkg.pd.Series)
    assert isinstance(qs[["name"]], DataTable)
    assert qs.all().count() == 3
    assert qs.none().count() == 0
    assert qs.filter().count() == 3
    assert qs.filter("age > 20").count() == 2
    assert qs.exclude().count() == 0
    assert qs.exclude("age > 20").count() == 1
    assert qs.order_by().count() == 3
    assert qs.order_by("age").first()["age"].iloc[0] == 19
    assert qs.order_by("?").count() == 3
    assert qs.distinct().count() == 3
    assert qs.distinct("name").to_df().columns.tolist() == ["name"]
    assert qs.only().count() == 3
    assert qs.exclude_fields().count() == 3
    assert qs.values("name").first() == {"name": "alice"}
    assert qs.values_list("name", flat=True).first() == "alice"
    with pytest.raises(ValueError):
        qs.values_list("name", "age", flat=True)
    assert qs.get(name="alice")["age"].iloc[0] == 19
    with pytest.raises(KeyError):
        qs.get(name="missing")
    with pytest.raises(ValueError):
        qs.get(active=True)
    assert qs.count() == 3
    assert qs.exists() is True
    assert qs.paginate(limit=2).count() == 2
    with pytest.raises(ValueError):
        qs.paginate(limit=-1)
    with pytest.raises(ValueError):
        qs.group_by()
    assert qs.aggregate() == {}
    assert qs.sum("age") == {"age": 74}
    assert qs.avg("age") == {"age": 24.666666666666668}
    assert qs.min("age") == {"age": 19}
    assert qs.max("age") == {"age": 31}
    assert qs.aggregate(total=("age", "sum")) == {"total": 74}
    with pytest.raises(KeyError):
        qs.sum("missing")
    with pytest.raises(KeyError):
        qs.aggregate(total=("missing", "sum"))

    assert qs.to_markdown_table().startswith("|")
    assert "<table" in qs.to_html_table()
    assert "alice" in qs.to_csv_table()
    assert qs.render_table("markdown").startswith("|")
    assert "<table" in qs.render_table("html")
    assert "alice" in qs.render_table("csv")

    # rich / excel branches are exercised with direct helpers in the next test.
    joined = qs.join(DataTable({"id": [1, 2, 3], "city": ["hn", "sg", "dn"]}), on="id")
    assert "city" in joined.to_df().columns
    assert "city" in qs.left_join(DataTable({"id": [1], "city": ["hn"]}), on="id").to_df().columns
    assert "city" in qs.inner_join(DataTable({"id": [1], "city": ["hn"]}), on="id").to_df().columns
    assert qs.anti_join(DataTable({"id": [1], "city": ["hn"]}), on="id").count() == 2
    assert qs.ensure_columns("missing").to_df().columns.tolist()[-1] == "missing"
    assert qs.coerce_types({"age": "int64"}).to_df()["age"].dtype.kind in {"i", "u"}
    assert qs.validate_schema({"name": {"nullable": False}}).equals(table)
    assert qs.auto_version("auto")._frame.is_auto_versioning() is True
    assert qs.disable_auto_version()._frame.is_auto_versioning() is False
    assert isinstance(qs.snapshot("snap"), dtpkg._ORMQuerySet)

    with pytest.raises(KeyError):
        qs.filter(missing=1)
    with pytest.raises(KeyError):
        qs.exclude(missing=1)
    assert qs.join(DataTable({"id": [1]}), on="id").count() == 1
    with pytest.raises(ValueError):
        qs.join(DataTable({"id": [1]}), on=[])
    with pytest.raises(ValueError):
        qs.join(DataTable({"id": [1]}), left_on="id")
    with pytest.raises(ValueError):
        qs.join(DataTable({"id": [1]}), left_on=["id"], right_on=["id", "x"])
    assert qs.ensure_columns().count() == 3
    with pytest.raises(KeyError):
        qs.coerce_types({"missing": "int64"})
    with pytest.raises(ValueError):
        qs.validate_schema({"name": {"choices": ["zzz"]}})

    group = qs.group_by("active")
    assert group.count().to_df().columns.tolist() == ["active", "count"]
    assert group.annotate().to_df().columns.tolist() == ["active", "count"]
    assert group.annotate(total=("age", "sum")).to_df().columns.tolist() == ["active", "total"]
    assert group.annotate(age="sum").to_df().columns.tolist() == ["active", "age"]
    with pytest.raises(KeyError):
        group.annotate(total=("missing", "sum"))
    with pytest.raises(TypeError):
        group.aggregate(age=lambda s: s.max())
    with pytest.raises(TypeError):
        group.aggregate(age=123)
    assert group.sum("age")["age"]
    assert group.avg("age")["age"]
    assert group.min("age")["age"]
    assert group.max("age")["age"]
    assert group.aggregate(age="sum")["age"]
    with pytest.raises(KeyError):
        group.sum("missing")
    with pytest.raises(KeyError):
        group.aggregate(total=("missing", "sum"))

    # direct queryset delegations to diff helpers
    assert qs.diff_columns_against(table)["column_order_changed"] is False
    assert qs.diff_added_rows(table).shape[0] == 0
    assert qs.diff_removed_rows(table).shape[0] == 0
    assert qs.diff_unchanged_rows(table).shape[0] == 3
    assert qs.diff_changed_rows(table).shape[0] == 0
    assert qs.diff_changed_cells(table).shape[0] == 0
    assert qs.diff_report(table)["has_changes"] is False


def test_render_excel_and_rich_branches(monkeypatch):
    table = DataTable({"name": ["alice", None], "age": [19, 24]})

    assert table.to_markdown_table().startswith("|")
    assert "<table" in table.to_html_table()
    assert "alice" in table.to_csv_table()
    assert "alice" in table.render_table("markdown")
    assert "<table" in table.render_table("html")
    assert "alice" in table.render_table("csv")

    rich_table = table.to_rich_table(title="demo", show_index=True)
    assert rich_table is not None
    assert table.render_table("rich") is not None

    fake_openpyxl = types.ModuleType("openpyxl")
    fake_styles = types.ModuleType("openpyxl.styles")

    class FakeFont:
        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs

    class FakePatternFill:
        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs

    class FakeCell:
        def __init__(self, value, column_letter):
            self.value = value
            self.column_letter = column_letter
            self.font = None
            self.fill = None

    class FakeWorksheet:
        def __init__(self):
            self.freeze_panes = None
            self._columns = [
                (FakeCell("name", "A"), FakeCell("alice", "A")),
                (FakeCell("age", "B"), FakeCell(19, "B")),
            ]
            self.column_dimensions = {"A": types.SimpleNamespace(width=None), "B": types.SimpleNamespace(width=None)}

        @property
        def columns(self):
            return self._columns

        def __getitem__(self, item):
            if item == 1:
                return [column[0] for column in self._columns]
            raise KeyError(item)

    class FakeWorkbook:
        def __init__(self):
            self.worksheet = FakeWorksheet()

        def __getitem__(self, key):
            return self.worksheet

        def save(self, path):
            Path(path).write_text("fake workbook", encoding="utf-8")

    def fake_load_workbook(path):
        return FakeWorkbook()

    fake_openpyxl.load_workbook = fake_load_workbook
    fake_styles.Font = FakeFont
    fake_styles.PatternFill = FakePatternFill
    monkeypatch.setitem(sys.modules, "openpyxl", fake_openpyxl)
    monkeypatch.setitem(sys.modules, "openpyxl.styles", fake_styles)

    def fake_to_excel(self, output, sheet_name="Sheet1", index=False, **kwargs):
        Path(output).write_text("excel", encoding="utf-8")

    monkeypatch.setattr(DataTable, "to_excel", fake_to_excel, raising=False)

    excel_path = Path("tests") / "_datatable_coverage.xlsx"
    try:
        output = table.to_excel_table(excel_path, style_header=True, autofit=True, freeze_header=True)
        assert output == excel_path
        assert excel_path.exists()
        assert table.render_table("excel", path=excel_path) == excel_path
    finally:
        if excel_path.exists():
            excel_path.unlink()


def test_testcase_mixin_branches(monkeypatch):
    class Sample(DataTable):
        pass

    sample = Sample({"x": [1]})
    Sample.setUpClass()
    Sample.addClassCleanup(lambda: None)

    entered = []

    class CM:
        def __enter__(self):
            entered.append("enter")
            return "entered"

        def __exit__(self, exc_type, exc, tb):
            entered.append("exit")

    assert Sample.enterClassContext(CM()) == "entered"
    with pytest.raises(RuntimeError):
        Sample.addClassCleanup(lambda: (_ for _ in ()).throw(RuntimeError("cleanup")))
        Sample.doClassCleanups()
    Sample.tearDownClass()
    assert entered == ["enter", "exit"]
    sample.setUp()
    sample.tearDown()
    sample.assertIsSubclass(Sample, DataTable)
    sample.assertNotIsSubclass(int, DataTable)
    with pytest.raises(AssertionError):
        sample.assertIsSubclass(int, DataTable)
    with pytest.raises(AssertionError):
        sample.assertNotIsSubclass(Sample, DataTable)
    with pytest.raises(AssertionError):
        sample.assertIsSubclass(123, DataTable)
    with pytest.raises(AssertionError):
        sample.assertNotIsSubclass(123, DataTable)

    monkeypatch.setattr(sample, "_testcase", lambda: types.SimpleNamespace())
    with sample.assertNoLogs("soigia.edge", level=logging.WARNING):
        pass



from soigia.tester import Tester


def test_tester_artifacts_roundtrip():
    tester = Tester()

    text_path = tester.write_text("sample.txt", "hello")
    assert text_path.exists()
    assert tester.read_text("sample.txt") == "hello"

    yaml_path = tester.write_yaml("config.yaml", {"name": "demo", "enabled": True})
    assert yaml_path.exists()
    payload = tester.read_yaml("config.yaml")
    assert payload["name"] == "demo"
    assert payload["enabled"] is True

    assert tester.exists("sample.txt")
    assert tester.root_dir().name == "soigia-sdk"
    tester.cleanup()



__version__ = "0.1.12"

__all__ = [
    "__version__",
    "Model",
    "fields",
    "setup",
    "use_storage",
    "config",
    "DataTable",
    "datatable",
    "df",
    "ui",
    "db",
    "pipeline",
    "Pipeline",
    "BasePipeline",
    "BaseStep",
    "FunctionStep",
    "PipelineContext",
    "PipelineStatus",
    "StepOutcome",
    "ConfigStore",
    "default_config",
    "load_config",
    "save_config",
    "resolve_config_path",
    "get_accounts",
    "get_account",
    "upsert_account",
    "delete_account",
    "DatabaseConnector",
    "db_connector",
    "from_sql",
    "from_mysql_db",
    "from_maria_db",
    "from_oracle_db",
    "from_postgres_db",
    "from_sqlite_db",
    "from_faker",
    "from_factory",
    "Tester",
]


def __getattr__(name):
    """Lazy import optional public helpers."""
    if name in {"Model", "fields", "setup", "use_storage"}:
        from .erp import Model, fields, setup, use_storage

        return {
            "Model": Model,
            "fields": fields,
            "setup": setup,
            "use_storage": use_storage,
        }[name]
    if name == "datatable":
        from importlib import import_module

        return import_module(".datatable", __name__)
    if name == "ui":
        from importlib import import_module

        return import_module(".ui", __name__)
    if name == "pipeline":
        from importlib import import_module

        return import_module(".pipeline", __name__)
    if name == "db":
        from importlib import import_module

        return import_module(".db", __name__)
    if name == "config":
        from importlib import import_module

        return import_module(".config", __name__)
    if name == "tester":
        from importlib import import_module

        return import_module(".tester", __name__)
    if name in {
        "Pipeline",
        "BasePipeline",
        "BaseStep",
        "FunctionStep",
        "PipelineContext",
        "PipelineStatus",
        "StepOutcome",
    }:
        from .pipeline import (
            BasePipeline,
            BaseStep,
            FunctionStep,
            Pipeline,
            PipelineContext,
            PipelineStatus,
            StepOutcome,
        )

        return {
            "Pipeline": Pipeline,
            "BasePipeline": BasePipeline,
            "BaseStep": BaseStep,
            "FunctionStep": FunctionStep,
            "PipelineContext": PipelineContext,
            "PipelineStatus": PipelineStatus,
            "StepOutcome": StepOutcome,
        }[name]
    if name in {"DataTable", "df"}:
        from .datatable import DataTable

        return DataTable
    if name in {
        "ConfigStore",
        "default_config",
        "load_config",
        "save_config",
        "resolve_config_path",
        "get_accounts",
        "get_account",
        "upsert_account",
        "delete_account",
    }:
        from .config import (
            ConfigStore,
            default_config,
            delete_account,
            get_account,
            get_accounts,
            load_config,
            resolve_config_path,
            save_config,
            upsert_account,
        )

        return {
            "ConfigStore": ConfigStore,
            "default_config": default_config,
            "load_config": load_config,
            "save_config": save_config,
            "resolve_config_path": resolve_config_path,
            "get_accounts": get_accounts,
            "get_account": get_account,
            "upsert_account": upsert_account,
            "delete_account": delete_account,
        }[name]
    if name in {
        "DatabaseConnector",
        "db_connector",
        "from_sql",
        "from_mysql_db",
        "from_maria_db",
        "from_oracle_db",
        "from_postgres_db",
        "from_sqlite_db",
        "from_faker",
        "from_factory",
    }:
        from .datatable import (
            DatabaseConnector,
            db_connector,
            from_factory,
            from_faker,
            from_maria_db,
            from_mysql_db,
            from_oracle_db,
            from_postgres_db,
            from_sql,
            from_sqlite_db,
        )

        return {
            "DatabaseConnector": DatabaseConnector,
            "db_connector": db_connector,
            "from_sql": from_sql,
            "from_mysql_db": from_mysql_db,
            "from_maria_db": from_maria_db,
            "from_oracle_db": from_oracle_db,
            "from_postgres_db": from_postgres_db,
            "from_sqlite_db": from_sqlite_db,
            "from_faker": from_faker,
            "from_factory": from_factory,
        }[name]
    if name == "Tester":
        from .tester import Tester

        return Tester
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



from __future__ import annotations

"""Runnable pipeline demo showing the current `DataTable` workflow.

This example is intentionally small but complete:
- `load_data()` loads a CSV into `self.datatable`
- `self.datatable` is used for all in-memory transforms
- `push_datatable()` writes into the pipeline-managed SQLite database
- `save_to_csv()` exports the current table
- `load_datatable()` reads the SQLite table back into `self.datatable`
"""

from pathlib import Path

from soigia.pipeline import Pipeline


DEMO_SOURCE_CSV = Path(__file__).with_name("demo_users.csv")
DEMO_UPSERT_CSV = Path(__file__).with_name("demo_upsert_users.csv")


class Shopee(Pipeline):
    _name = "Shopee"
    _inherit = ["sale", "purchase"]
    _data = ["product", "user", "customer"]
    cli_show_datatable_default = True

    stages = ["normalize", "save_users", "export_users", "reload_users"]

    def load_data(self):
        self.datatable = self.datatable.load_from_csv(DEMO_SOURCE_CSV)

    def normalize(self):
        self.datatable = self.datatable.copy()
        self.datatable["name"] = self.datatable["name"].astype(str).str.strip().str.title()
        self.datatable["status"] = self.datatable["status"].astype(str).str.lower()
        self.datatable["net_amount"] = (self.datatable["amount"] * 0.98).round(2)

    def save_users(self):
        self.push_datatable("users", self.datatable, mode="replace")

    def export_users(self):
        self.datatable.save_to_csv("data/demo/users.csv")

    def reload_users(self):
        self.datatable = self.load_datatable("users").sort_values("id").reset_index(drop=True)


class ShopeeUpsertDemo(Pipeline):
    stages = ["seed_users", "apply_changes", "reload_users", "export_vip_user"]

    def load_data(self):
        self.datatable = self.datatable.load_from_csv(DEMO_UPSERT_CSV)

    def seed_users(self):
        self.push_datatable("users", self.datatable, mode="replace")

    def apply_changes(self):
        # Upsert keeps id=2 and inserts id=3.
        self.upsert_rows(
            "users",
            [
                {"id": 2, "name": "bob", "status": "vip", "amount": 250.0},
                {"id": 3, "name": "carol", "status": "active", "amount": 300.0},
            ],
            key_columns=["id"],
            mode="upsert",
        )

    def reload_users(self):
        self.datatable = self.load_datatable("users").sort_values("id").reset_index(drop=True)

    def export_vip_user(self):
        # This is the chaining style you asked for:
        # filter(...) -> get(...) -> save_to_csv(...)
        self.datatable.filter(status="vip").get(id=2).save_to_csv("data/demo/user_2_vip.csv")


if __name__ == "__main__":
    Shopee.cli()



"""Self-contained ORM layer for Soi Gia.

This module keeps the ORM core in one file on purpose:
- declarative fields and model metadata
- registry and environment objects
- query layer
- memory, pandas, and sqlite backends
- session and unit-of-work helpers
"""

from __future__ import annotations

import copy
import inspect
import json
import sqlite3
from collections import OrderedDict
from dataclasses import dataclass, field
from datetime import date, datetime
from pathlib import Path
from types import SimpleNamespace
from typing import Any, Callable, Iterable, Iterator

try:
    import pandas as pd
except Exception:  # pragma: no cover - pandas is expected in this repo
    pd = None  # type: ignore[assignment]

__all__ = [
    "Expression",
    "Condition",
    "Field",
    "String",
    "Integer",
    "Float",
    "Boolean",
    "Date",
    "DateTime",
    "Text",
    "ForeignKey",
    "OneToMany",
    "ManyToMany",
    "ModelOptions",
    "Registry",
    "BackendAdapter",
    "MemoryBackend",
    "PandasBackend",
    "SQLiteBackend",
    "Repository",
    "RecordSet",
    "QuerySet",
    "Manager",
    "Env",
    "Session",
    "UnitOfWork",
    "Model",
    "fields",
    "default_registry",
    "default_env",
]


def _call_maybe(func: Callable[..., Any], instance: Any = None) -> Any:
    try:
        return func()
    except TypeError:
        if instance is None:
            return func(instance)
        return func(instance)


def _normalize_related_value(value: Any) -> Any:
    if value is None:
        return None
    if hasattr(value, "id"):
        return getattr(value, "id")
    if isinstance(value, RecordSet):
        return value.ids
    return value


def _json_default(value: Any) -> Any:
    if isinstance(value, (date, datetime)):
        return value.isoformat()
    if isinstance(value, Path):
        return str(value)
    if hasattr(value, "id"):
        return getattr(value, "id")
    if isinstance(value, set):
        return list(value)
    raise TypeError(f"Object of type {type(value).__name__} is not JSON serializable")


LOOKUPS = {
    "eq",
    "ne",
    "gt",
    "gte",
    "lt",
    "lte",
    "in",
    "contains",
    "icontains",
    "startswith",
    "istartswith",
    "endswith",
    "iendswith",
    "isnull",
}

LOOKUP_ALIASES = {
    "=": "eq",
    "==": "eq",
    "!=": "ne",
    "<>": "ne",
}


class Expression:
    def evaluate(self, row: dict[str, Any], model_cls: type["Model"], env: "Env") -> bool:
        raise NotImplementedError


@dataclass(slots=True)
class Condition(Expression):
    field_name: str
    lookup: str = "eq"
    value: Any = None
    negated: bool = False

    def evaluate(self, row: dict[str, Any], model_cls: type["Model"], env: "Env") -> bool:
        field = model_cls._meta.fields.get(self.field_name)
        actual = row.get(self.field_name)
        expected = _normalize_related_value(self.value)

        if field is not None:
            actual = field.from_storage(actual, None)
            if isinstance(field, ForeignKey):
                actual = _normalize_related_value(actual)
        else:
            actual = _normalize_related_value(actual)

        result = self._compare(actual, expected)
        return not result if self.negated else result

    def _compare(self, actual: Any, expected: Any) -> bool:
        lookup = LOOKUP_ALIASES.get(self.lookup, self.lookup)
        if lookup == "eq":
            return actual == expected
        if lookup == "ne":
            return actual != expected
        if lookup == "gt":
            return actual is not None and expected is not None and actual > expected
        if lookup == "gte":
            return actual is not None and expected is not None and actual >= expected
        if lookup == "lt":
            return actual is not None and expected is not None and actual < expected
        if lookup == "lte":
            return actual is not None and expected is not None and actual <= expected
        if lookup == "in":
            return actual in (expected or [])
        if lookup == "contains":
            return actual is not None and expected is not None and expected in actual
        if lookup == "icontains":
            return actual is not None and expected is not None and str(expected).lower() in str(actual).lower()
        if lookup == "startswith":
            return actual is not None and expected is not None and str(actual).startswith(str(expected))
        if lookup == "istartswith":
            return actual is not None and expected is not None and str(actual).lower().startswith(str(expected).lower())
        if lookup == "endswith":
            return actual is not None and expected is not None and str(actual).endswith(str(expected))
        if lookup == "iendswith":
            return actual is not None and expected is not None and str(actual).lower().endswith(str(expected).lower())
        if lookup == "isnull":
            return (actual is None) == bool(expected)
        raise ValueError(f"unsupported lookup: {lookup}")


class Field:
    python_type: type[Any] = Any

    def __init__(
        self,
        default: Any = None,
        *,
        required: bool = False,
        nullable: bool = True,
        compute: str | Callable[..., Any] | None = None,
        store: bool = True,
        read_only: bool = False,
        help_text: str | None = None,
    ) -> None:
        self.default = default
        self.required = required
        self.nullable = nullable
        self.compute = compute
        self.store = store
        self.read_only = read_only
        self.help_text = help_text
        self.name = ""
        self.model: type["Model"] | None = None

    def clone(self) -> "Field":
        return copy.copy(self)

    def __set_name__(self, owner: type["Model"], name: str) -> None:
        self.name = name
        self.model = owner

    def get_default(self, instance: "Model" | None = None) -> Any:
        if callable(self.default):
            return _call_maybe(self.default, instance)
        return copy.deepcopy(self.default)

    def validate(self, value: Any) -> Any:
        if value is None:
            if self.required and not self.nullable:
                raise ValueError(f"field {self.name!r} cannot be null")
            return None
        return value

    def to_storage(self, value: Any) -> Any:
        return value

    def from_storage(self, value: Any, instance: "Model" | None = None) -> Any:
        return value

    def compute_value(self, instance: "Model") -> Any:
        if self.compute is None:
            return None
        target = self.compute
        if isinstance(target, str):
            target = getattr(instance, target)
        return _call_maybe(target, instance)

    def __get__(self, instance: "Model" | None, owner: type["Model"]) -> Any:
        if instance is None:
            return self
        if self.compute is not None:
            value = self.compute_value(instance)
            if self.store:
                instance._values[self.name] = self.to_storage(value)
            return value
        if self.name in instance._values:
            value = instance._values[self.name]
            return self.from_storage(value, instance)
        return self.get_default(instance)

    def __set__(self, instance: "Model", value: Any) -> None:
        if self.read_only:
            raise AttributeError(f"field {self.name!r} is read-only")
        validated = self.validate(value)
        instance._values[self.name] = self.to_storage(validated)


class String(Field):
    python_type = str

    def validate(self, value: Any) -> Any:
        value = super().validate(value)
        if value is None:
            return None
        return str(value)


class Text(String):
    pass


class Integer(Field):
    python_type = int

    def validate(self, value: Any) -> Any:
        value = super().validate(value)
        if value is None:
            return None
        return int(value)


class Float(Field):
    python_type = float

    def validate(self, value: Any) -> Any:
        value = super().validate(value)
        if value is None:
            return None
        return float(value)


class Boolean(Field):
    python_type = bool

    def validate(self, value: Any) -> Any:
        value = super().validate(value)
        if value is None:
            return None
        if isinstance(value, str):
            return value.strip().lower() in {"1", "true", "yes", "y", "on"}
        return bool(value)


class Date(Field):
    python_type = date

    def validate(self, value: Any) -> Any:
        value = super().validate(value)
        if value is None:
            return None
        if isinstance(value, date) and not isinstance(value, datetime):
            return value
        if isinstance(value, str):
            return date.fromisoformat(value)
        raise TypeError(f"field {self.name!r} expects a date")

    def to_storage(self, value: Any) -> Any:
        return value.isoformat() if isinstance(value, date) else value

    def from_storage(self, value: Any, instance: "Model" | None = None) -> Any:
        if value is None or isinstance(value, date):
            return value
        return date.fromisoformat(str(value))


class DateTime(Field):
    python_type = datetime

    def validate(self, value: Any) -> Any:
        value = super().validate(value)
        if value is None:
            return None
        if isinstance(value, datetime):
            return value
        if isinstance(value, str):
            return datetime.fromisoformat(value)
        raise TypeError(f"field {self.name!r} expects a datetime")

    def to_storage(self, value: Any) -> Any:
        return value.isoformat() if isinstance(value, datetime) else value

    def from_storage(self, value: Any, instance: "Model" | None = None) -> Any:
        if value is None or isinstance(value, datetime):
            return value
        return datetime.fromisoformat(str(value))


class ForeignKey(Field):
    def __init__(self, comodel_name: str, default: Any = None, **kwargs: Any) -> None:
        super().__init__(default=default, **kwargs)
        self.comodel_name = comodel_name

    def validate(self, value: Any) -> Any:
        value = super().validate(value)
        if value is None:
            return None
        return _normalize_related_value(value)

    def __get__(self, instance: "Model" | None, owner: type["Model"]) -> Any:
        if instance is None:
            return self
        raw = instance._values.get(self.name)
        if raw is None:
            return self.get_default(instance)
        env = instance._env or default_env
        if env is None:
            return raw
        return env[self.comodel_name].browse(raw)


class OneToMany(Field):
    def __init__(self, comodel_name: str, inverse_name: str, **kwargs: Any) -> None:
        super().__init__(default=list, store=False, read_only=True, **kwargs)
        self.comodel_name = comodel_name
        self.inverse_name = inverse_name

    def __get__(self, instance: "Model" | None, owner: type["Model"]) -> Any:
        if instance is None:
            return self
        env = instance._env or default_env
        if env is None or instance.id is None:
            return RecordSet(owner=None, records=[])
        return env[self.comodel_name].search([(self.inverse_name, "=", instance.id)])

    def __set__(self, instance: "Model", value: Any) -> None:  # pragma: no cover - read only relation
        raise AttributeError(f"field {self.name!r} is read-only")


class ManyToMany(Field):
    def __init__(self, comodel_name: str, default: Any = None, **kwargs: Any) -> None:
        super().__init__(default=default or list, **kwargs)
        self.comodel_name = comodel_name

    def validate(self, value: Any) -> Any:
        value = super().validate(value)
        if value is None:
            return []
        if isinstance(value, RecordSet):
            return value.ids
        if isinstance(value, (list, tuple, set)):
            return [_normalize_related_value(item) for item in value]
        return [_normalize_related_value(value)]

    def __get__(self, instance: "Model" | None, owner: type["Model"]) -> Any:
        if instance is None:
            return self
        raw = instance._values.get(self.name)
        if raw is None:
            return self.get_default(instance) or []
        env = instance._env or default_env
        if env is None:
            return raw
        return env[self.comodel_name].browse(raw)


@dataclass(slots=True)
class ModelOptions:
    model_name: str
    table_name: str
    abstract: bool = False
    fields: OrderedDict[str, Field] = field(default_factory=OrderedDict)
    field_order: list[str] = field(default_factory=list)


class Registry:
    def __init__(self) -> None:
        self.models: dict[str, type["Model"]] = {}
        self.env: Env | None = None

    def bind_env(self, env: "Env") -> None:
        self.env = env
        for model in self.models.values():
            model._env = env
            env.backend.ensure_model(model)

    def register(self, model_cls: type["Model"]) -> type["Model"]:
        model_name = model_cls._meta.model_name
        if not model_name:
            raise ValueError(f"model {model_cls.__name__} must define _name")
        self.models[model_name] = model_cls
        model_cls._registry = self
        model_cls._env = self.env
        if self.env is not None:
            self.env.backend.ensure_model(model_cls)
        return model_cls

    def extend(self, extension_cls: type["Model"]) -> type["Model"]:
        inherit = extension_cls._inherit
        target_names = [inherit] if isinstance(inherit, str) else list(inherit or [])
        if not target_names:
            raise ValueError(f"extension {extension_cls.__name__} must define _inherit")
        targets: list[type["Model"]] = []
        for target_name in target_names:
            target = self.models.get(target_name)
            if target is None:
                raise KeyError(f"unknown model to inherit: {target_name}")
            targets.append(target)
        for target in targets:
            self._merge_extension(target, extension_cls)
        return targets[0]

    def _merge_extension(self, target: type["Model"], extension_cls: type["Model"]) -> None:
        for name, attr in extension_cls.__dict__.items():
            if name.startswith("_"):
                continue
            if isinstance(attr, Field):
                cloned = attr.clone()
                setattr(target, name, cloned)
                cloned.__set_name__(target, name)
                target._meta.fields[name] = cloned
                if name not in target._meta.field_order:
                    target._meta.field_order.append(name)
                continue
            if callable(attr):
                setattr(target, name, attr)


default_registry = Registry()


class BackendAdapter:
    name = "base"

    def ensure_model(self, model_cls: type["Model"]) -> None:
        return None

    def insert(self, model_cls: type["Model"], row: dict[str, Any]) -> dict[str, Any]:
        raise NotImplementedError

    def update(self, model_cls: type["Model"], record_id: int, row: dict[str, Any]) -> dict[str, Any]:
        raise NotImplementedError

    def delete(self, model_cls: type["Model"], record_id: int) -> None:
        raise NotImplementedError

    def get(self, model_cls: type["Model"], record_id: int) -> dict[str, Any] | None:
        raise NotImplementedError

    def all(self, model_cls: type["Model"]) -> list[dict[str, Any]]:
        raise NotImplementedError

    def search(self, model_cls: type["Model"], predicate: Callable[[dict[str, Any]], bool]) -> list[dict[str, Any]]:
        return [row for row in self.all(model_cls) if predicate(row)]


class MemoryBackend(BackendAdapter):
    name = "memory"

    def __init__(self) -> None:
        self._tables: dict[str, dict[int, dict[str, Any]]] = {}
        self._sequences: dict[str, int] = {}

    def ensure_model(self, model_cls: type["Model"]) -> None:
        self._tables.setdefault(model_cls._meta.model_name, {})
        self._sequences.setdefault(model_cls._meta.model_name, 0)

    def _next_id(self, model_cls: type["Model"]) -> int:
        name = model_cls._meta.model_name
        self._sequences[name] = self._sequences.get(name, 0) + 1
        return self._sequences[name]

    def insert(self, model_cls: type["Model"], row: dict[str, Any]) -> dict[str, Any]:
        self.ensure_model(model_cls)
        if row.get("id") is None:
            row["id"] = self._next_id(model_cls)
        self._tables[model_cls._meta.model_name][int(row["id"])] = copy.deepcopy(row)
        return copy.deepcopy(row)

    def update(self, model_cls: type["Model"], record_id: int, row: dict[str, Any]) -> dict[str, Any]:
        self.ensure_model(model_cls)
        current = self._tables[model_cls._meta.model_name].get(int(record_id))
        if current is None:
            raise KeyError(f"record {record_id} not found")
        current.update(copy.deepcopy(row))
        current["id"] = int(record_id)
        return copy.deepcopy(current)

    def delete(self, model_cls: type["Model"], record_id: int) -> None:
        self.ensure_model(model_cls)
        self._tables[model_cls._meta.model_name].pop(int(record_id), None)

    def get(self, model_cls: type["Model"], record_id: int) -> dict[str, Any] | None:
        self.ensure_model(model_cls)
        row = self._tables[model_cls._meta.model_name].get(int(record_id))
        return copy.deepcopy(row) if row is not None else None

    def all(self, model_cls: type["Model"]) -> list[dict[str, Any]]:
        self.ensure_model(model_cls)
        rows = list(self._tables[model_cls._meta.model_name].values())
        return [copy.deepcopy(row) for row in sorted(rows, key=lambda item: item["id"])]


class PandasBackend(BackendAdapter):
    name = "pandas"

    def __init__(self) -> None:
        if pd is None:  # pragma: no cover - pandas is expected in this repo
            raise RuntimeError("pandas is required for PandasBackend")
        self._tables: dict[str, Any] = {}
        self._sequences: dict[str, int] = {}

    def ensure_model(self, model_cls: type["Model"]) -> None:
        name = model_cls._meta.model_name
        if name not in self._tables:
            self._tables[name] = pd.DataFrame(columns=["id", *model_cls._meta.field_order])
            self._sequences[name] = 0

    def _next_id(self, model_cls: type["Model"]) -> int:
        name = model_cls._meta.model_name
        self._sequences[name] = self._sequences.get(name, 0) + 1
        return self._sequences[name]

    def insert(self, model_cls: type["Model"], row: dict[str, Any]) -> dict[str, Any]:
        self.ensure_model(model_cls)
        row = copy.deepcopy(row)
        if row.get("id") is None:
            row["id"] = self._next_id(model_cls)
        frame = self._tables[model_cls._meta.model_name]
        if frame.empty:
            self._tables[model_cls._meta.model_name] = pd.DataFrame([row], columns=frame.columns)
        else:
            self._tables[model_cls._meta.model_name] = pd.concat([frame, pd.DataFrame([row])], ignore_index=True)
        return copy.deepcopy(row)

    def update(self, model_cls: type["Model"], record_id: int, row: dict[str, Any]) -> dict[str, Any]:
        self.ensure_model(model_cls)
        frame = self._tables[model_cls._meta.model_name]
        mask = frame["id"] == record_id
        if not mask.any():
            raise KeyError(f"record {record_id} not found")
        for key, value in row.items():
            frame.loc[mask, key] = value
        self._tables[model_cls._meta.model_name] = frame
        return copy.deepcopy(frame[mask].iloc[0].to_dict())

    def delete(self, model_cls: type["Model"], record_id: int) -> None:
        self.ensure_model(model_cls)
        frame = self._tables[model_cls._meta.model_name]
        self._tables[model_cls._meta.model_name] = frame[frame["id"] != record_id].reset_index(drop=True)

    def get(self, model_cls: type["Model"], record_id: int) -> dict[str, Any] | None:
        self.ensure_model(model_cls)
        frame = self._tables[model_cls._meta.model_name]
        match = frame[frame["id"] == record_id]
        if match.empty:
            return None
        return copy.deepcopy(match.iloc[0].to_dict())

    def all(self, model_cls: type["Model"]) -> list[dict[str, Any]]:
        self.ensure_model(model_cls)
        frame = self._tables[model_cls._meta.model_name]
        return [copy.deepcopy(row) for row in frame.sort_values("id").to_dict(orient="records")]


class SQLiteBackend(BackendAdapter):
    name = "sqlite"

    def __init__(self, path: str | Path = ":memory:") -> None:
        self.path = str(path)
        self._connection = sqlite3.connect(self.path)
        self._connection.row_factory = sqlite3.Row

    def ensure_model(self, model_cls: type["Model"]) -> None:
        table = model_cls._meta.table_name
        self._connection.execute(
            f'CREATE TABLE IF NOT EXISTS "{table}" (id INTEGER PRIMARY KEY AUTOINCREMENT, payload TEXT NOT NULL)'
        )
        self._connection.commit()

    def insert(self, model_cls: type["Model"], row: dict[str, Any]) -> dict[str, Any]:
        self.ensure_model(model_cls)
        cur = self._connection.execute(
            f'INSERT INTO "{model_cls._meta.table_name}" (payload) VALUES (?)',
            (json.dumps(row, default=_json_default),),
        )
        self._connection.commit()
        row = copy.deepcopy(row)
        row["id"] = cur.lastrowid
        self._connection.execute(
            f'UPDATE "{model_cls._meta.table_name}" SET payload = ? WHERE id = ?',
            (json.dumps(row, default=_json_default), row["id"]),
        )
        self._connection.commit()
        return row

    def update(self, model_cls: type["Model"], record_id: int, row: dict[str, Any]) -> dict[str, Any]:
        current = self.get(model_cls, record_id)
        if current is None:
            raise KeyError(f"record {record_id} not found")
        current.update(copy.deepcopy(row))
        current["id"] = int(record_id)
        self._connection.execute(
            f'UPDATE "{model_cls._meta.table_name}" SET payload = ? WHERE id = ?',
            (json.dumps(current, default=_json_default), int(record_id)),
        )
        self._connection.commit()
        return current

    def delete(self, model_cls: type["Model"], record_id: int) -> None:
        self.ensure_model(model_cls)
        self._connection.execute(f'DELETE FROM "{model_cls._meta.table_name}" WHERE id = ?', (int(record_id),))
        self._connection.commit()

    def get(self, model_cls: type["Model"], record_id: int) -> dict[str, Any] | None:
        self.ensure_model(model_cls)
        cur = self._connection.execute(
            f'SELECT payload FROM "{model_cls._meta.table_name}" WHERE id = ?',
            (int(record_id),),
        )
        row = cur.fetchone()
        if row is None:
            return None
        return copy.deepcopy(json.loads(row["payload"]))

    def all(self, model_cls: type["Model"]) -> list[dict[str, Any]]:
        self.ensure_model(model_cls)
        cur = self._connection.execute(f'SELECT payload FROM "{model_cls._meta.table_name}" ORDER BY id')
        return [copy.deepcopy(json.loads(row["payload"])) for row in cur.fetchall()]


class Repository:
    def __init__(self, model_cls: type["Model"], backend: BackendAdapter, env: "Env") -> None:
        self.model_cls = model_cls
        self.backend = backend
        self.env = env

    def create(self, values: dict[str, Any]) -> dict[str, Any]:
        row = self.model_cls._prepare_row(values, env=self.env)
        return self.backend.insert(self.model_cls, row)

    def update(self, record_id: int, values: dict[str, Any]) -> dict[str, Any]:
        row = self.model_cls._prepare_row(values, env=self.env, partial=True)
        return self.backend.update(self.model_cls, record_id, row)

    def delete(self, record_id: int) -> None:
        self.backend.delete(self.model_cls, record_id)

    def get(self, record_id: int) -> dict[str, Any] | None:
        return self.backend.get(self.model_cls, record_id)

    def all(self) -> list[dict[str, Any]]:
        return self.backend.all(self.model_cls)

    def search(self, conditions: Iterable[Condition]) -> list[dict[str, Any]]:
        condition_list = list(conditions)

        def predicate(row: dict[str, Any]) -> bool:
            return all(condition.evaluate(row, self.model_cls, self.env) for condition in condition_list)

        return self.backend.search(self.model_cls, predicate)


class RecordSet:
    def __init__(self, owner: type["Model"] | None, records: list["Model"] | None = None) -> None:
        self.owner = owner
        self._records = list(records or [])

    def __iter__(self) -> Iterator["Model"]:
        return iter(self._records)

    def __len__(self) -> int:
        return len(self._records)

    def __bool__(self) -> bool:
        return bool(self._records)

    def __getitem__(self, item: int | slice) -> "Model" | "RecordSet":
        if isinstance(item, slice):
            return RecordSet(self.owner, self._records[item])
        return self._records[item]

    def __repr__(self) -> str:
        model = self.owner.__name__ if self.owner else "RecordSet"
        return f"<{model}RecordSet {self.ids}>"

    def __getattr__(self, name: str) -> Any:
        if not self._records:
            raise AttributeError(name)
        attr = getattr(self._records[0], name)
        if callable(attr):

            def caller(*args: Any, **kwargs: Any) -> list[Any]:
                return [getattr(record, name)(*args, **kwargs) for record in self._records]

            return caller
        if len(self._records) == 1:
            return attr
        return [getattr(record, name) for record in self._records]

    @property
    def ids(self) -> list[int]:
        return [record.id for record in self._records if record.id is not None]

    def first(self) -> "Model" | None:
        return self._records[0] if self._records else None

    def all(self) -> list["Model"]:
        return list(self._records)

    def mapped(self, name: str) -> list[Any]:
        return [getattr(record, name) for record in self._records]

    def filtered(self, predicate: Callable[["Model"], bool]) -> "RecordSet":
        return RecordSet(self.owner, [record for record in self._records if predicate(record)])

    def update(self, **values: Any) -> "RecordSet":
        for record in self._records:
            for key, value in values.items():
                setattr(record, key, value)
            record.save()
        return self

    def delete(self) -> None:
        for record in list(self._records):
            record.delete()
        self._records.clear()


class QuerySet:
    def __init__(
        self,
        model_cls: type["Model"],
        env: "Env",
        *,
        conditions: list[Condition] | None = None,
        orderings: list[str] | None = None,
        limit_value: int | None = None,
    ) -> None:
        self.model_cls = model_cls
        self.env = env
        self.conditions = list(conditions or [])
        self.orderings = list(orderings or [])
        self.limit_value = limit_value

    def _clone(self, **kwargs: Any) -> "QuerySet":
        return QuerySet(
            self.model_cls,
            self.env,
            conditions=kwargs.get("conditions", self.conditions),
            orderings=kwargs.get("orderings", self.orderings),
            limit_value=kwargs.get("limit_value", self.limit_value),
        )

    def _repository(self) -> Repository:
        return Repository(self.model_cls, self.env.backend, self.env)

    def _evaluate_rows(self) -> list[dict[str, Any]]:
        rows = self._repository().search(self.conditions) if self.conditions else self._repository().all()
        if self.orderings:
            rows = self._apply_ordering(rows)
        if self.limit_value is not None:
            rows = rows[: self.limit_value]
        return rows

    def _apply_ordering(self, rows: list[dict[str, Any]]) -> list[dict[str, Any]]:
        ordered = list(rows)
        for ordering in reversed(self.orderings):
            reverse = ordering.startswith("-")
            field_name = ordering[1:] if reverse else ordering

            def sort_key(row: dict[str, Any]) -> tuple[bool, Any]:
                value = row.get(field_name)
                return (value is None, value)

            ordered.sort(key=sort_key, reverse=reverse)
        return ordered

    def _records(self) -> list["Model"]:
        return [self.model_cls._from_row(row, env=self.env) for row in self._evaluate_rows()]

    def filter(self, *domain: Any, **lookups: Any) -> "QuerySet":
        conditions = self.conditions + _build_conditions(*domain, **lookups)
        return self._clone(conditions=conditions)

    def exclude(self, *domain: Any, **lookups: Any) -> "QuerySet":
        conditions = self.conditions + [
            Condition(condition.field_name, condition.lookup, condition.value, not condition.negated)
            for condition in _build_conditions(*domain, **lookups)
        ]
        return self._clone(conditions=conditions)

    def search(self, domain: Iterable[tuple[str, str, Any]] | None = None, **lookups: Any) -> "RecordSet":
        return self.filter(*(list(domain or [])), **lookups).all()

    def order_by(self, *fields: str) -> "QuerySet":
        return self._clone(orderings=self.orderings + list(fields))

    def limit(self, value: int) -> "QuerySet":
        return self._clone(limit_value=int(value))

    def all(self) -> RecordSet:
        return RecordSet(self.model_cls, self._records())

    def first(self) -> "Model" | None:
        return self.all().first()

    def create(self, values: dict[str, Any] | None = None, **kwargs: Any) -> "Model" | RecordSet:
        payload = dict(values or {})
        payload.update(kwargs)
        return self.model_cls.create(payload, env=self.env)

    def update(self, **values: Any) -> RecordSet:
        records = self.all()
        records.update(**values)
        return records

    def delete(self) -> None:
        self.all().delete()

    def browse(self, ids: int | Iterable[int]) -> "Model" | RecordSet | None:
        return self.model_cls.browse(ids, env=self.env)

    def __iter__(self) -> Iterator["Model"]:
        return iter(self.all())

    def __len__(self) -> int:
        return len(self.all())


class Manager:
    def __init__(self, model_cls: type["Model"] | None = None, env: "Env" | None = None) -> None:
        self.model_cls = model_cls
        self.env = env

    def __get__(self, instance: "Model" | None, owner: type["Model"]) -> "Manager":
        env = owner._env or default_env
        return Manager(owner, env)

    def get_queryset(self) -> QuerySet:
        if self.model_cls is None or self.env is None:
            raise RuntimeError("manager is not bound to a model")
        return QuerySet(self.model_cls, self.env)

    def filter(self, *domain: Any, **lookups: Any) -> QuerySet:
        return self.get_queryset().filter(*domain, **lookups)

    def exclude(self, *domain: Any, **lookups: Any) -> QuerySet:
        return self.get_queryset().exclude(*domain, **lookups)

    def order_by(self, *fields: str) -> QuerySet:
        return self.get_queryset().order_by(*fields)

    def limit(self, value: int) -> QuerySet:
        return self.get_queryset().limit(value)

    def all(self) -> RecordSet:
        return self.get_queryset().all()

    def first(self) -> "Model" | None:
        return self.get_queryset().first()

    def create(self, values: dict[str, Any] | None = None, **kwargs: Any) -> "Model" | RecordSet:
        return self.get_queryset().create(values, **kwargs)

    def search(self, domain: Iterable[tuple[str, str, Any]] | None = None, **lookups: Any) -> RecordSet:
        return self.get_queryset().search(domain, **lookups)

    def browse(self, ids: int | Iterable[int]) -> "Model" | RecordSet | None:
        return self.get_queryset().browse(ids)


class ModelProxy:
    def __init__(self, model_cls: type["Model"], env: "Env") -> None:
        self.model_cls = model_cls
        self.env = env

    @property
    def objects(self) -> Manager:
        return Manager(self.model_cls, self.env)

    def create(self, values: dict[str, Any] | None = None, **kwargs: Any) -> "Model" | RecordSet:
        return self.model_cls.create(values, env=self.env, **kwargs)

    def search(self, domain: Iterable[tuple[str, str, Any]] | None = None, **lookups: Any) -> RecordSet:
        return self.model_cls.search(domain, env=self.env, **lookups)

    def browse(self, ids: int | Iterable[int]) -> "Model" | RecordSet | None:
        return self.model_cls.browse(ids, env=self.env)

    def all(self) -> RecordSet:
        return self.model_cls.all(env=self.env)

    def __getattr__(self, name: str) -> Any:
        manager = self.objects
        if hasattr(manager, name):
            return getattr(manager, name)
        return getattr(self.model_cls, name)


class Env:
    def __init__(self, registry: Registry | None = None, backend: BackendAdapter | None = None) -> None:
        self.registry = registry or default_registry
        self.backend = backend or MemoryBackend()
        self.registry.bind_env(self)

    def __getitem__(self, model_name: str) -> ModelProxy:
        return ModelProxy(self.registry.models[model_name], self)

    def repository(self, model_cls: type["Model"]) -> Repository:
        return Repository(model_cls, self.backend, self)


class UnitOfWork:
    def __init__(self, env: Env) -> None:
        self.env = env
        self.new: list[Model] = []
        self.dirty: list[Model] = []
        self.deleted: list[Model] = []

    def register_new(self, record: "Model") -> "Model":
        self.new.append(record)
        return record

    def register_dirty(self, record: "Model") -> "Model":
        self.dirty.append(record)
        return record

    def register_deleted(self, record: "Model") -> "Model":
        self.deleted.append(record)
        return record

    def commit(self) -> None:
        for record in self.new:
            record.save()
        for record in self.dirty:
            record.save()
        for record in self.deleted:
            record.delete()
        self.rollback()

    def rollback(self) -> None:
        self.new.clear()
        self.dirty.clear()
        self.deleted.clear()


class Session:
    def __init__(self, env: Env | None = None) -> None:
        self.env = env or default_env
        self.uow = UnitOfWork(self.env)

    def __enter__(self) -> "Session":
        return self

    def __exit__(self, exc_type, exc, tb) -> None:
        if exc is None:
            self.commit()
        else:
            self.rollback()

    def add(self, record: "Model") -> "Model":
        return self.uow.register_new(record)

    def dirty(self, record: "Model") -> "Model":
        return self.uow.register_dirty(record)

    def delete(self, record: "Model") -> "Model":
        return self.uow.register_deleted(record)

    def commit(self) -> None:
        self.uow.commit()

    def rollback(self) -> None:
        self.uow.rollback()


default_env: Env


def _build_conditions(*domain: Any, **lookups: Any) -> list[Condition]:
    conditions: list[Condition] = []
    for item in domain:
        if isinstance(item, Condition):
            conditions.append(item)
            continue
        if isinstance(item, (list, tuple)) and len(item) == 3:
            field_name, lookup, value = item
            conditions.append(Condition(str(field_name), str(lookup), value))
            continue
        if isinstance(item, (list, tuple)):
            conditions.extend(_build_conditions(*item))
            continue
        raise TypeError(f"unsupported domain item: {item!r}")

    for key, value in lookups.items():
        parts = key.split("__")
        if parts[-1] in LOOKUPS:
            field_name = "__".join(parts[:-1])
            lookup = parts[-1]
        else:
            field_name = key
            lookup = "eq"
        lookup = LOOKUP_ALIASES.get(lookup, lookup)
        conditions.append(Condition(field_name, lookup, value))
    return conditions


class ModelMeta(type):
    def __new__(mcls, name: str, bases: tuple[type[Any], ...], attrs: dict[str, Any]) -> type[Any]:
        attrs = dict(attrs)
        inherited_fields: OrderedDict[str, Field] = OrderedDict()
        for base in bases:
            base_meta = getattr(base, "_meta", None)
            if base_meta is None:
                continue
            for field_name, field_obj in base_meta.fields.items():
                if field_name in attrs:
                    continue
                inherited_fields[field_name] = field_obj.clone()
        attrs.update(inherited_fields)
        cls = super().__new__(mcls, name, bases, attrs)

        collected_fields: OrderedDict[str, Field] = OrderedDict()
        for base in bases:
            base_meta = getattr(base, "_meta", None)
            if base_meta is None:
                continue
            for field_name, field_obj in base_meta.fields.items():
                collected_fields[field_name] = field_obj.clone()
        for attr_name, attr_value in cls.__dict__.items():
            if isinstance(attr_value, Field):
                collected_fields[attr_name] = attr_value

        model_name = attrs.get("_name")
        inherit = attrs.get("_inherit")
        abstract = bool(attrs.get("_abstract", False) or name == "Model")
        table_name = str(model_name or name.lower())
        cls._meta = ModelOptions(
            model_name=str(model_name or ""),
            table_name=table_name,
            abstract=abstract,
            fields=collected_fields,
            field_order=list(collected_fields.keys()),
        )
        cls._registry = default_registry
        cls._env = None
        cls.objects = Manager()

        if inherit and not model_name:
            default_registry.extend(cls)
            return cls

        if not abstract:
            if not model_name:
                cls._meta.model_name = name.lower()
                cls._meta.table_name = name.lower()
            default_registry.register(cls)
        return cls


class Model(metaclass=ModelMeta):
    _abstract = True
    _name: str = ""
    _inherit: str | list[str] | None = None
    _registry: Registry
    _env: Env | None
    _meta: ModelOptions
    objects: Manager

    def __init__(self, env: Env | None = None, id: int | None = None, **values: Any) -> None:
        self._env = env or self.__class__._env or default_env
        self._values: dict[str, Any] = {}
        self._persisted = False
        self.id = id
        for key, value in values.items():
            setattr(self, key, value)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id!r}>"

    @classmethod
    def _get_env(cls, env: Env | None = None) -> Env:
        resolved = env or cls._env or default_env
        if resolved is None:
            raise RuntimeError("no ORM environment is available")
        return resolved

    @classmethod
    def _repository(cls, env: Env | None = None) -> Repository:
        resolved_env = cls._get_env(env)
        return Repository(cls, resolved_env.backend, resolved_env)

    @classmethod
    def _prepare_row(cls, values: dict[str, Any], env: Env, *, partial: bool = False) -> dict[str, Any]:
        row: dict[str, Any] = {}
        provided = dict(values)
        for field_name, field_obj in cls._meta.fields.items():
            if not field_obj.store:
                continue
            if field_name in provided:
                raw_value = provided.pop(field_name)
            else:
                if partial:
                    continue
                raw_value = field_obj.get_default(None)
            validated = field_obj.validate(raw_value)
            row[field_name] = field_obj.to_storage(validated)
        if provided and not partial:
            unknown = ", ".join(sorted(provided))
            raise AttributeError(f"unknown fields for {cls.__name__}: {unknown}")
        if partial:
            for key, value in provided.items():
                field_obj = cls._meta.fields.get(key)
                if field_obj is None or not field_obj.store:
                    continue
                row[key] = field_obj.to_storage(field_obj.validate(value))
        if "id" in values and values["id"] is not None:
            row["id"] = int(values["id"])
        return row

    @classmethod
    def _from_row(cls, row: dict[str, Any], env: Env) -> "Model":
        obj = cls.__new__(cls)
        obj._env = env
        obj._values = {}
        obj._persisted = True
        obj.id = row.get("id")
        for field_name, field_obj in cls._meta.fields.items():
            if field_name not in row:
                continue
            obj._values[field_name] = field_obj.from_storage(row[field_name], obj)
        return obj

    @classmethod
    def create(cls, values: dict[str, Any] | None = None, *, env: Env | None = None, **kwargs: Any) -> "Model" | RecordSet:
        payload = dict(values or {})
        payload.update(kwargs)
        resolved_env = cls._get_env(env)
        row = cls._repository(resolved_env).create(payload)
        return cls._from_row(row, resolved_env)

    @classmethod
    def browse(cls, ids: int | Iterable[int], *, env: Env | None = None) -> "Model" | RecordSet | None:
        resolved_env = cls._get_env(env)
        repo = cls._repository(resolved_env)
        if isinstance(ids, Iterable) and not isinstance(ids, (str, bytes)):
            records = []
            for record_id in ids:
                row = repo.get(int(record_id))
                if row is not None:
                    records.append(cls._from_row(row, resolved_env))
            return RecordSet(cls, records)
        row = repo.get(int(ids))
        return cls._from_row(row, resolved_env) if row is not None else None

    @classmethod
    def search(cls, domain: Iterable[tuple[str, str, Any]] | None = None, *, env: Env | None = None, **lookups: Any) -> RecordSet:
        resolved_env = cls._get_env(env)
        return QuerySet(cls, resolved_env).search(domain, **lookups)

    @classmethod
    def all(cls, *, env: Env | None = None) -> RecordSet:
        return QuerySet(cls, cls._get_env(env)).all()

    def to_dict(self, *, raw: bool = False) -> dict[str, Any]:
        payload: dict[str, Any] = {"id": self.id}
        for field_name, field_obj in self._meta.fields.items():
            if not field_obj.store:
                continue
            value = self._values.get(field_name)
            payload[field_name] = value if raw else field_obj.from_storage(value, self)
        return payload

    def save(self) -> "Model":
        env = self._env or default_env
        repo = self._repository(env)
        payload = self.to_dict(raw=True)
        payload.pop("id", None)
        if self.id is None:
            row = repo.create(payload)
            self.id = row["id"]
            self._values.update({k: v for k, v in row.items() if k != "id"})
            self._persisted = True
            return self
        row = repo.update(int(self.id), payload)
        self._values.update({k: v for k, v in row.items() if k != "id"})
        self._persisted = True
        return self

    def delete(self) -> None:
        if self.id is None:
            return
        repo = self._repository(self._env or default_env)
        repo.delete(int(self.id))
        self.id = None
        self._persisted = False

    def refresh(self) -> "Model":
        if self.id is None:
            raise RuntimeError("cannot refresh an unsaved record")
        row = self._repository(self._env or default_env).get(int(self.id))
        if row is None:
            raise KeyError(f"record {self.id} not found")
        fresh = self.__class__._from_row(row, self._env or default_env)
        self._values = fresh._values
        self._persisted = True
        return self

    def copy(self, **overrides: Any) -> "Model":
        payload = self.to_dict(raw=False)
        payload.pop("id", None)
        payload.update(overrides)
        return self.__class__.create(payload, env=self._env or default_env)

    def __setattr__(self, name: str, value: Any) -> None:
        if name in {"_env", "_values", "_persisted", "id"}:
            object.__setattr__(self, name, value)
            return
        meta = getattr(self.__class__, "_meta", None)
        if meta is not None and name in meta.fields:
            meta.fields[name].__set__(self, value)
            return
        object.__setattr__(self, name, value)


fields = SimpleNamespace(
    Field=Field,
    String=String,
    Integer=Integer,
    Float=Float,
    Boolean=Boolean,
    Date=Date,
    DateTime=DateTime,
    Text=Text,
    ForeignKey=ForeignKey,
    OneToMany=OneToMany,
    ManyToMany=ManyToMany,
)


default_env = Env(default_registry, MemoryBackend())



import pytest

from soigia.pipeline import BasePipeline, Pipeline, PipelineStatus
from soigia.datatable import DataTable


def test_pipeline_decorator_and_hooks():
    pipeline = BasePipeline(name="demo")
    events = []

    @pipeline.before_run
    def before(ctx):
        events.append(("before_run", ctx.run_id))

    @pipeline.after_run
    def after(ctx):
        events.append(("after_run", ctx.run_id))

    @pipeline.step(name="double", retries=1)
    def double(data, ctx):
        events.append(("step", ctx.current_step, ctx.current_attempt))
        return data * 2

    outcome = pipeline.run(10)

    assert outcome.status == PipelineStatus.SUCCESS
    assert outcome.data == 20
    assert events[0][0] == "before_run"
    assert events[-1][0] == "after_run"
    assert events[1] == ("step", "double", 1)


def test_pipeline_condition_and_skip():
    pipeline = BasePipeline(name="demo")

    @pipeline.step(name="skip_me", condition=lambda data, ctx: False)
    def skipped(data, ctx):  # pragma: no cover - this should not run
        return data + 1

    outcome = pipeline.run(3)

    assert outcome.status == PipelineStatus.SUCCESS
    assert outcome.data == 3
    assert pipeline.steps[0].name == "skip_me"


def test_pipeline_subclass_auto_registers_decorated_methods():
    class DemoPipeline(Pipeline):
        stages = ["double", "add_three"]

        def load_data(self):
            return 4

        def double(self):
            self.datatable = self.datatable * 2
            return self.datatable

        def add_three(self):
            self.datatable = self.datatable + 3
            return self.datatable

    pipeline = DemoPipeline(name="demo")

    assert [step.name for step in pipeline.steps] == ["double", "add_three"]
    outcome = pipeline.run()
    assert outcome.status == PipelineStatus.SUCCESS
    assert isinstance(outcome.data, DataTable)
    assert outcome.data.to_rows() == [{"value": 11}]


def test_pipeline_exposes_shared_datatable_and_model():
    class DemoPipeline(Pipeline):
        stages = ["capture", "finalize"]

        def load_data(self):
            return 4

        def capture(self):
            assert isinstance(self.datatable, DataTable)
            assert self.datatable.to_rows() == [{"value": 4}]
            self.model.seen = getattr(self.model, "seen", 0) + 1
            self.datatable = self.datatable.assign(value=self.datatable["value"] + 1)
            return self.datatable

        def finalize(self):
            assert self.model.seen == 1
            assert self.datatable.to_rows() == [{"value": 5}]
            self.datatable = self.datatable.assign(value=self.datatable["value"] * 2)
            return self.datatable

    pipeline = DemoPipeline(name="demo")
    outcome = pipeline.run()

    assert outcome.status == PipelineStatus.SUCCESS
    assert isinstance(outcome.data, DataTable)
    assert outcome.data.to_rows() == [{"value": 10}]
    assert isinstance(pipeline.datatable, DataTable)
    assert pipeline.datatable.to_rows() == [{"value": 10}]
    assert pipeline.model.seen == 1


def test_pipeline_stage_methods_can_mutate_datatable_without_returning():
    class DemoPipeline(Pipeline):
        stages = ["capture", "finalize"]

        def load_data(self):
            return DataTable({"value": [4]})

        def capture(self):
            assert isinstance(self.datatable, DataTable)
            self.datatable = self.datatable.assign(value=self.datatable["value"] + 1)

        def finalize(self):
            assert isinstance(self.datatable, DataTable)
            self.datatable = self.datatable.assign(value=self.datatable["value"] * 2)

    pipeline = DemoPipeline(name="demo")
    outcome = pipeline.run()

    assert outcome.status == PipelineStatus.SUCCESS
    assert isinstance(outcome.data, DataTable)
    assert outcome.data.to_rows() == [{"value": 10}]
    assert isinstance(pipeline.datatable, DataTable)
    assert pipeline.datatable.to_rows() == [{"value": 10}]


def test_pipeline_datatable_supports_chainable_row_query_and_save(monkeypatch):
    from pathlib import Path

    sandbox = Path.cwd() / ".test_artifacts" / "pipeline_datatable_chain"
    sandbox.mkdir(parents=True, exist_ok=True)
    monkeypatch.chdir(sandbox)

    class DemoPipeline(Pipeline):
        stages = ["export"]

        def load_data(self):
            return DataTable(
                [
                    {"id": 1, "name": "alice", "status": "new"},
                    {"id": 2, "name": "bob", "status": "vip"},
                    {"id": 3, "name": "carol", "status": "active"},
                ]
            )

        def export(self):
            self.datatable.filter(status="vip").get(id=2).save_to_csv("exports/vip_user.csv")

    pipeline = DemoPipeline(name="demo")
    result = pipeline.run()

    assert result.success
    output = sandbox / "exports" / "vip_user.csv"
    assert output.exists()
    chain_result = pipeline.datatable.filter(status="vip").get(id=2)
    assert isinstance(chain_result, DataTable)
    assert chain_result.to_rows() == [{"id": 2, "name": "bob", "status": "vip"}]


def test_pipeline_helper_methods():
    import pandas as pd
    from pathlib import Path

    class DemoPipeline(Pipeline):
        stages = ["prepare", "finalize"]

        def load_data(self):
            import pandas as pd

            return pd.DataFrame({"name": ["alice"], "amount": [3]})

        def prepare(self):
            assert self.has_columns("name", "amount")
            self.snapshot_df("before")
            self.put("note", "ready")
            self.datatable = self.datatable.assign(amount2=self.datatable["amount"] * 2)
            return self.datatable

        def finalize(self):
            assert self.get("note") == "ready"
            self.restore_df("before")
            assert "amount2" not in self.datatable.columns
            self.datatable = self.datatable.assign(done=True)
            return self.datatable

    pipeline = DemoPipeline(name="demo")
    result = pipeline.run()

    assert result.status == PipelineStatus.SUCCESS
    assert result.rows == 1
    assert result.csv_path is not None
    assert result.db_path is not None
    assert result.parquet_path is not None
    assert result.summary_path is not None
    assert result.duration_seconds is not None
    assert pipeline.datatable.equals(pd.DataFrame({"name": ["alice"], "amount": [3], "done": [True]}))
    dump_dir = Path.cwd() / "data" / "DemoPipeline"
    assert (dump_dir / "01_prepare.csv").exists()
    assert (dump_dir / "02_finalize.csv").exists()
    summary_path = dump_dir / "pipeline_summary.md"
    summary_text = summary_path.read_text(encoding="utf-8", errors="replace")
    assert "# Pipeline Run Summary" in summary_text
    assert "| # | Step | Status | Rows | Added Columns | Removed Columns |" in summary_text
    assert "| 01 | prepare | success |" in summary_text
    assert "| 02 | finalize | success |" in summary_text
    assert "## Model" in summary_text


def test_pipeline_exposes_default_sqlite_db_and_upsert(monkeypatch):
    import pandas as pd
    from pathlib import Path

    sandbox = Path.cwd() / ".test_artifacts" / "pipeline_db_upsert"
    sandbox.mkdir(parents=True, exist_ok=True)
    monkeypatch.chdir(sandbox)

    class DemoPipeline(Pipeline):
        stages = ["seed", "upsert", "read_back"]

        def load_data(self):
            return pd.DataFrame({"id": [1, 2], "value": ["a", "b"]})

        def seed(self):
            self.db.set("items", self.datatable)
            return self.datatable

        def upsert(self):
            self.upsert_rows("items", [{"id": 2, "value": "b-updated"}, {"id": 3, "value": "c"}])
            return self.datatable

        def read_back(self):
            self.datatable = self.db.get("items").sort_values("id").reset_index(drop=True)
            return self.datatable

    pipeline = DemoPipeline(name="demo", config={"db_upsert_keys": {"items": ["id"]}})

    assert pipeline.db_path.exists()
    assert pipeline.table_exists("soigia_meta")
    assert "soigia_meta" in pipeline.list_tables()

    result = pipeline.run()

    assert result.success
    assert list(pipeline.datatable["id"]) == [1, 2, 3]
    assert list(pipeline.datatable["value"]) == ["a", "b-updated", "c"]
    meta = pipeline.db.read_df("soigia_meta").set_index("key")["value"].to_dict()
    assert meta["pipeline_class"] == "DemoPipeline"
    assert meta["pipeline_name"] == "demo"
    assert meta["db_path"] == str(pipeline.db_path)


def test_pipeline_respects_env_default_db_path(monkeypatch):
    from pathlib import Path
    import pandas as pd

    sandbox = Path.cwd() / ".test_artifacts" / "pipeline_env_default_db"
    sandbox.mkdir(parents=True, exist_ok=True)
    monkeypatch.chdir(sandbox)
    monkeypatch.setenv("SOIGIA_DB_PATH", str(sandbox / "custom" / "pipeline.sqlite3"))

    class DemoPipeline(Pipeline):
        stages = ["seed"]

        def load_data(self):
            return pd.DataFrame({"id": [1], "value": ["a"]})

        def seed(self):
            self.db.set("items", self.datatable)
            return self.datatable

    pipeline = DemoPipeline(name="demo")

    assert pipeline.db_path == (sandbox / "custom" / "pipeline.sqlite3").resolve()
    result = pipeline.run()
    assert result.success
    assert pipeline.db_path.exists()


def test_pipeline_supports_delete_insert_mode(monkeypatch):
    import pandas as pd
    from pathlib import Path

    sandbox = Path.cwd() / ".test_artifacts" / "pipeline_db_delete_insert"
    sandbox.mkdir(parents=True, exist_ok=True)
    monkeypatch.chdir(sandbox)

    class DemoPipeline(Pipeline):
        stages = ["seed", "sync", "read_back"]

        def load_data(self):
            return pd.DataFrame({"id": [1, 2], "value": ["a", "b"]})

        def seed(self):
            self.db.set("items", self.datatable)
            return self.datatable

        def sync(self):
            self.upsert_rows(
                "items",
                [{"id": 2, "value": "b-replaced"}, {"id": 3, "value": "c"}],
                mode="delete-insert",
            )
            return self.datatable

        def read_back(self):
            self.datatable = self.db.get("items").sort_values("id").reset_index(drop=True)
            return self.datatable

    pipeline = DemoPipeline(name="demo", config={"db_upsert_keys": {"items": ["id"]}})
    result = pipeline.run()

    assert result.success
    assert list(pipeline.datatable["id"]) == [1, 2, 3]
    assert list(pipeline.datatable["value"]) == ["a", "b-replaced", "c"]
    assert pipeline.db.table_exists("items")


def test_pipeline_supports_insert_mode_and_aliases(monkeypatch):
    import pandas as pd
    from pathlib import Path

    sandbox = Path.cwd() / ".test_artifacts" / "pipeline_db_insert"
    sandbox.mkdir(parents=True, exist_ok=True)
    monkeypatch.chdir(sandbox)

    class DemoPipeline(Pipeline):
        stages = ["seed", "append", "read_back"]

        def load_data(self):
            return pd.DataFrame({"id": [1], "value": ["a"]})

        def seed(self):
            self.db.set("items", self.datatable)
            return self.datatable

        def append(self):
            self.db.upsert_rows(
                "items",
                [{"id": 2, "value": "b"}, {"id": 3, "value": "c"}],
                mode="insert",
            )
            return self.datatable

        def read_back(self):
            self.datatable = self.db.get("items").sort_values("id").reset_index(drop=True)
            return self.datatable

    pipeline = DemoPipeline(name="demo")
    result = pipeline.run()

    assert result.success
    assert list(pipeline.datatable["id"]) == [1, 2, 3]
    assert list(pipeline.datatable["value"]) == ["a", "b", "c"]
    assert pipeline.table_exists("items")
    assert "items" in pipeline.list_tables()


def test_pipeline_syncs_data_df_to_db_and_back(monkeypatch):
    import pandas as pd
    from pathlib import Path

    sandbox = Path.cwd() / ".test_artifacts" / "pipeline_df_sync"
    sandbox.mkdir(parents=True, exist_ok=True)
    monkeypatch.chdir(sandbox)

    class DemoPipeline(Pipeline):
        stages = ["push", "pull"]

        def load_data(self):
            return pd.DataFrame({"id": [1, 2], "value": ["a", "b"]})

        def push(self):
            self.sync_data_df("items")
            self.from_df("audit", self.datatable)
            return self.datatable

        def pull(self):
            self.load_data_df("items")
            return self.datatable

    pipeline = DemoPipeline(name="demo")
    result = pipeline.run()

    assert result.success
    assert pipeline.df.equals(pd.DataFrame({"id": [1, 2], "value": ["a", "b"]}))
    assert pipeline.to_df("items").equals(pd.DataFrame({"id": [1, 2], "value": ["a", "b"]}))
    assert pipeline.to_df("audit").equals(pd.DataFrame({"id": [1, 2], "value": ["a", "b"]}))
    assert pipeline.table_exists("items")
    assert pipeline.table_exists("audit")
    assert pipeline.get_data("items").equals(pd.DataFrame({"id": [1, 2], "value": ["a", "b"]}))
    assert pipeline._data["items"].equals(pd.DataFrame({"id": [1, 2], "value": ["a", "b"]}))


def test_pipeline_envs_supports_quick_sqlite_reads(monkeypatch):
    import pandas as pd
    from pathlib import Path

    sandbox = Path.cwd() / ".test_artifacts" / "pipeline_envs"
    sandbox.mkdir(parents=True, exist_ok=True)
    monkeypatch.chdir(sandbox)

    class DemoPipeline(Pipeline):
        stages = ["seed", "query"]

        def load_data(self):
            return pd.DataFrame({"id": [1, 2, 3], "value": ["a", "b", "c"]})

        def seed(self):
            self.sync_data_df("orders")
            return self.datatable

        def query(self):
            rows = self.envs.orders.filter(self.envs.orders.id > 1).to_df()
            self.set_data("filtered_orders", rows)
            return self.datatable

    pipeline = DemoPipeline(name="demo")
    result = pipeline.run()

    assert result.success
    assert pipeline.get_data("filtered_orders").equals(pd.DataFrame({"id": [2, 3], "value": ["b", "c"]}))
    assert pipeline.envs.orders.filter(id__gt=1).to_df().equals(pd.DataFrame({"id": [2, 3], "value": ["b", "c"]}))


def test_pipeline_supports_datatable_alias(monkeypatch):
    import pandas as pd
    from pathlib import Path

    sandbox = Path.cwd() / ".test_artifacts" / "pipeline_datatable_alias"
    sandbox.mkdir(parents=True, exist_ok=True)
    monkeypatch.chdir(sandbox)

    class DemoPipeline(Pipeline):
        stages = ["seed", "load_back"]

        def load_data(self):
            return pd.DataFrame({"id": [1], "value": ["a"]})

        def seed(self):
            self.datatable = self.datatable.assign(value=self.datatable["value"].str.upper())
            self.sync_datatable("items")
            return self.datatable

        def load_back(self):
            self.load_datatable("items")
            return self.datatable

    pipeline = DemoPipeline(name="demo")
    result = pipeline.run()

    assert result.success
    assert pipeline.get_datatable().equals(pd.DataFrame({"id": [1], "value": ["A"]}))
    assert pipeline.datatable.equals(pd.DataFrame({"id": [1], "value": ["A"]}))


def test_pipeline_datatable_property_set_and_get():
    import pandas as pd

    pipeline = Pipeline(name="demo")
    frame = pd.DataFrame({"id": [1], "value": ["a"]})

    pipeline.datatable = frame

    assert pipeline.datatable.equals(frame)
    assert pipeline.get_datatable().equals(frame)

    replacement = pd.DataFrame({"id": [2], "value": ["b"]})
    assert pipeline.set_datatable(replacement).equals(replacement)
    assert pipeline.datatable.equals(replacement)


def test_pipeline_datatable_coerces_tabular_values_to_datatable():
    import pandas as pd

    pipeline = Pipeline(name="demo")

    pipeline.datatable = pd.DataFrame({"id": [1], "value": ["a"]})
    assert isinstance(pipeline.datatable, DataTable)

    pipeline.datatable = [{"id": 2, "value": "b"}]
    assert isinstance(pipeline.datatable, DataTable)

    pipeline.datatable = {"id": [3], "value": ["c"]}
    assert isinstance(pipeline.datatable, DataTable)


def test_pipeline_datatable_allows_non_tabular_values():
    pipeline = Pipeline(name="demo")

    pipeline.datatable = 123
    assert isinstance(pipeline.datatable, DataTable)
    assert pipeline.datatable.to_rows() == [{"value": 123}]
    assert isinstance(pipeline.datatable, DataTable)
    assert pipeline.datatable.to_rows() == [{"value": 123}]


def test_pipeline_db_upsert_update_columns_and_invalid_mode(monkeypatch):
    import pandas as pd
    from pathlib import Path

    sandbox = Path.cwd() / ".test_artifacts" / "pipeline_upsert_update_columns"
    sandbox.mkdir(parents=True, exist_ok=True)
    monkeypatch.chdir(sandbox)

    class DemoPipeline(Pipeline):
        stages = ["seed", "sync", "read_back"]

        def load_data(self):
            return pd.DataFrame({"id": [1], "name": ["alice"], "status": ["new"]})

        def seed(self):
            self.db.set("items", self.datatable)
            return self.datatable

        def sync(self):
            self.db.upsert(
                "items",
                pd.DataFrame({"id": [1], "name": ["alice-updated"], "status": ["ignored"]}),
                key_columns=["id"],
                update_columns=["name"],
            )
            return self.datatable

        def read_back(self):
            self.datatable = self.load_datatable("items")
            return self.datatable

    pipeline = DemoPipeline(name="demo")
    result = pipeline.run()

    assert result.success
    assert list(pipeline.datatable["name"]) == ["alice-updated"]
    assert list(pipeline.datatable["status"]) == ["new"]

    with pytest.raises(ValueError):
        pipeline.db.upsert("items", pd.DataFrame({"id": [2]}), mode="bad-mode")


def test_pipeline_push_and_pull_aliases(monkeypatch):
    import pandas as pd
    from pathlib import Path

    sandbox = Path.cwd() / ".test_artifacts" / "pipeline_push_pull_aliases"
    sandbox.mkdir(parents=True, exist_ok=True)
    monkeypatch.chdir(sandbox)

    pipeline = Pipeline(name="demo")
    payload = pd.DataFrame({"id": [1, 2], "value": ["a", "b"]})

    pipeline.push("items", payload)
    pulled = pipeline.pull("items")

    assert pulled.equals(payload)
    assert pipeline.datatable.equals(payload)
    assert pipeline.db.get("items").equals(payload)


def test_pipeline_db_delete_insert_creates_missing_table(monkeypatch):
    import pandas as pd
    from pathlib import Path

    sandbox = Path.cwd() / ".test_artifacts" / "pipeline_delete_insert_new_table"
    sandbox.mkdir(parents=True, exist_ok=True)
    monkeypatch.chdir(sandbox)

    class DemoPipeline(Pipeline):
        stages = ["sync", "read_back"]

        def load_data(self):
            return pd.DataFrame({"id": [1, 2], "value": ["a", "b"]})

        def sync(self):
            self.db.upsert("items", self.datatable, mode="delete-insert", key_columns=["id"])
            return self.datatable

        def read_back(self):
            self.datatable = self.load_datatable("items").sort_values("id").reset_index(drop=True)
            return self.datatable

    pipeline = DemoPipeline(name="demo")
    result = pipeline.run()

    assert result.success
    assert pipeline.datatable.equals(pd.DataFrame({"id": [1, 2], "value": ["a", "b"]}))
    assert pipeline.table_exists("items")


def test_pipeline_before_and_after_methods_are_called():
    class DemoPipeline(Pipeline):
        stages = ["step_one"]

        def load_data(self):
            return 0

        def before_run(self):
            self.model.events = ["before"]

        def after_run(self):
            self.model.events.append("after")

        def step_one(self):
            self.model.events.append("step")
            self.datatable = 1
            return self.datatable

    pipeline = DemoPipeline(name="demo")
    outcome = pipeline.run()

    assert outcome.status == PipelineStatus.SUCCESS
    assert pipeline.model.events == ["before", "step", "after"]


def test_pipeline_module_entrypoint_import():
    from soigia.pipeline import BasePipeline as PublicBasePipeline

    assert PublicBasePipeline is BasePipeline


def test_pipeline_exposes_source_path():
    from pathlib import Path

    class PathPipeline(Pipeline):
        stages = []

    pipeline = PathPipeline(name="demo")

    assert pipeline.path == Path(__file__).resolve()


def test_pipeline_display_datatable_returns_rendered_text(monkeypatch):
    from pathlib import Path

    sandbox = Path.cwd() / ".test_artifacts" / "pipeline_display_datatable"
    sandbox.mkdir(parents=True, exist_ok=True)
    monkeypatch.chdir(sandbox)

    class DemoPipeline(Pipeline):
        stages = []

        def load_data(self):
            return DataTable({"id": [1], "name": ["alice"]})

    pipeline = DemoPipeline(name="demo")
    pipeline.datatable = DataTable({"id": [1], "name": ["alice"]})
    text = pipeline.display_datatable()

    assert isinstance(text, str)
    assert "alice" in text
    assert "id" in text.lower()


def test_pipeline_logging_datatable_writes_to_log(monkeypatch):
    from pathlib import Path

    sandbox = Path.cwd() / ".test_artifacts" / "pipeline_logging_datatable"
    sandbox.mkdir(parents=True, exist_ok=True)
    monkeypatch.chdir(sandbox)

    class DemoPipeline(Pipeline):
        stages = []

        def load_data(self):
            return DataTable({"id": [1], "name": ["alice"]})

    pipeline = DemoPipeline(name="demo")
    pipeline.datatable = DataTable({"id": [1], "name": ["alice"]})
    text = pipeline.logging_datatable()

    log_path = Path("logs") / "DemoPipeline.log"
    assert text
    assert log_path.exists()
    content = log_path.read_text(encoding="utf-8", errors="replace")
    assert "alice" in content
    assert "id" in content.lower()


def test_pipeline_cli_runs_with_optional_overrides(monkeypatch, capsys):
    from pathlib import Path

    sandbox = Path.cwd() / ".test_artifacts" / "pipeline_cli"
    sandbox.mkdir(parents=True, exist_ok=True)
    monkeypatch.chdir(sandbox)

    class DemoPipeline(Pipeline):
        stages = ["normalize"]

        def load_data(self):
            return DataTable({"id": [1], "name": ["alice"]})

        def normalize(self):
            if self.config.get("uppercase"):
                self.datatable["name"] = self.datatable["name"].str.upper()

    outcome = DemoPipeline.cli(["--name", "cli-demo", "--set", "uppercase=true", "--step", "normalize", "--show-datatable"])
    captured = capsys.readouterr()

    assert outcome.status == PipelineStatus.SUCCESS
    assert outcome.data.to_rows() == [{"id": 1, "name": "ALICE"}]
    assert "ALICE" in captured.out


def test_pipeline_loaders_support_data_csv_and_yaml():
    import pandas as pd
    import yaml
    from pathlib import Path

    class LoaderPipeline(Pipeline):
        stages = ["normalize"]

        def load_data(self):
            return [{"name": " alice ", "amount": "12.5"}, {"name": "bob", "amount": "7"}]

        def normalize(self):
            self.logging("loaders-test-marker")
            df = pd.DataFrame(self.require_df())
            df["name"] = df["name"].astype(str).str.strip()
            df["amount"] = pd.to_numeric(df["amount"], errors="coerce")
            self.datatable = df
            return self.datatable

    data_pipeline = LoaderPipeline(name="data")
    data_result = data_pipeline.run()
    assert data_result.success
    assert data_pipeline.df.shape == (2, 2)
    dump_dir = Path.cwd() / "data" / "LoaderPipeline"
    assert (dump_dir / "01_normalize.csv").exists()
    log_path = Path.cwd() / "logs" / "LoaderPipeline.log"
    assert log_path.exists()
    assert "loaders-test-marker" in log_path.read_text(encoding="utf-8", errors="replace")
    summary_text = (dump_dir / "pipeline_summary.md").read_text(encoding="utf-8", errors="replace")
    assert "# Pipeline Run Summary" in summary_text
    assert "| 01 | normalize | success |" in summary_text

    artifact_root = Path.cwd() / ".test_artifacts"
    artifact_root.mkdir(exist_ok=True)

    csv_path = artifact_root / "sales-loader.csv"
    pd.DataFrame([{"name": "alice", "amount": 12.5}, {"name": "bob", "amount": 7.0}]).to_csv(csv_path, index=False)

    class CsvPipeline(Pipeline):
        stages = ["normalize"]

        def load_data(self):
            return self.load_csv(csv_path)

        def normalize(self):
            self.datatable["name"] = self.datatable["name"].str.upper()
            return self.datatable

    csv_pipeline = CsvPipeline(name="csv")
    csv_result = csv_pipeline.run()
    assert csv_result.success
    assert list(csv_pipeline.df["name"]) == ["ALICE", "BOB"]
    assert (Path.cwd() / "data" / "CsvPipeline" / "01_normalize.csv").exists()

    yaml_path = artifact_root / "sales-loader.yaml"
    yaml_payload = [{"name": "carol", "amount": 9.5}, {"name": "dave", "amount": 3.25}]
    yaml_path.write_text(
        yaml.safe_dump(yaml_payload, sort_keys=False, allow_unicode=True),
        encoding="utf-8",
    )

    class YamlPipeline(Pipeline):
        stages = ["normalize"]

        def load_data(self):
            payload = self.load_yaml(yaml_path)
            self.datatable = pd.DataFrame(payload)
            return self.datatable

        def normalize(self):
            self.datatable["amount"] = pd.to_numeric(self.datatable["amount"], errors="coerce")
            return self.datatable

    yaml_pipeline = YamlPipeline(name="yaml")
    yaml_result = yaml_pipeline.run()
    assert yaml_result.success
    assert yaml_pipeline.df.shape == (2, 2)
    assert (Path.cwd() / "data" / "YamlPipeline" / "01_normalize.csv").exists()


def test_pipeline_supports_explicit_csv_and_parquet_helpers():
    import pandas as pd
    from pathlib import Path

    artifact_root = Path.cwd() / ".test_artifacts" / "file_helpers"
    artifact_root.mkdir(parents=True, exist_ok=True)

    source_df = pd.DataFrame({"name": ["alice", "bob"], "amount": [12.5, 7.0]})
    csv_path = artifact_root / "source.csv"
    parquet_path = artifact_root / "source.parquet"
    source_df.to_csv(csv_path, index=False)

    class FileHelperPipeline(Pipeline):
        stages = ["load", "save"]

        def load_data(self):
            return self.load_data_from_csv(csv_path)

        def load(self):
            self.datatable["amount"] = pd.to_numeric(self.datatable["amount"], errors="coerce")
            return self.datatable

        def save(self):
            self.save_data_to_csv(artifact_root / "output.csv")
            self.save_data_to_parquet(parquet_path)
            return self.datatable

    pipeline = FileHelperPipeline(name="file-helper")
    result = pipeline.run()

    assert result.success
    assert (artifact_root / "output.csv").exists()
    assert parquet_path.exists()

    loaded = pipeline.load_data_from_parquet(parquet_path)
    assert loaded.equals(pd.DataFrame({"name": ["alice", "bob"], "amount": [12.5, 7.0]}))


def test_pipeline_pipelines_alias_still_works():
    class AliasPipeline(Pipeline):
        pipelines = ["double"]

        def load_data(self):
            return 2

        def double(self):
            self.datatable = self.datatable * 2
            return self.datatable

    pipeline = AliasPipeline(name="alias")
    outcome = pipeline.run()
    assert outcome.success
    assert isinstance(outcome.data, DataTable)
    assert outcome.data.to_rows() == [{"value": 4}]


def test_pipeline_protected_hooks_cannot_be_overridden():
    with pytest.raises(TypeError):

        class BadPreprocessPipeline(Pipeline):
            def preprocess_data(self):  # pragma: no cover - defined only to trigger the guard
                return self.datatable

    with pytest.raises(TypeError):

        class BadPostprocessPipeline(Pipeline):
            def postprocess_data(self):  # pragma: no cover - defined only to trigger the guard
                return self.datatable

    with pytest.raises(TypeError):

        class BadSavePipeline(Pipeline):
            def save_data_to_db(self):  # pragma: no cover - defined only to trigger the guard
                return self.datatable

    with pytest.raises(TypeError):

        class BadSaveOutputsPipeline(Pipeline):
            def save_outputs(self):  # pragma: no cover - defined only to trigger the guard
                return self.datatable


def test_pipeline_defaults_to_empty_dataframe_when_load_data_not_overridden():
    class EmptyPipeline(Pipeline):
        stages = []

    pipeline = EmptyPipeline(name="empty")
    result = pipeline.run()

    assert result.success
    assert result.rows == 0
    assert hasattr(pipeline.datatable, "empty")
    assert pipeline.datatable.empty
    assert result.csv_path is not None
    assert result.db_path is None
    assert result.parquet_path is None


def test_pipeline_uses_load_data_return_value():
    class ReturnPipeline(Pipeline):
        stages = ["double"]

        def load_data(self):
            return DataTable({"value": [3]})

        def double(self):
            self.datatable = self.datatable * 2
            return self.datatable

    pipeline = ReturnPipeline(name="return")
    result = pipeline.run()

    assert result.success
    assert isinstance(pipeline.datatable, DataTable)
    assert pipeline.datatable.to_rows() == [{"value": 6}]


def test_pipeline_loads_yaml_config_into_namespace():
    import yaml
    from pathlib import Path

    config_dir = Path(__file__).resolve().parent
    config_path = config_dir / "config.yaml"
    config_path.write_text(
        yaml.safe_dump({"factor": 4, "label": "demo", "nested": {"enabled": True}}, sort_keys=False),
        encoding="utf-8",
    )

    class YamlConfigPipeline(Pipeline):
        stages = ["scale"]

        def load_data(self):
            return DataTable({"value": [3]})

        def scale(self):
            self.datatable = self.datatable * self.config.factor
            self.model.label = self.config.label
            self.model.enabled = self.config.nested["enabled"]
            return self.datatable

    pipeline = YamlConfigPipeline(name="yaml-config")
    result = pipeline.run()

    assert result.success
    assert isinstance(pipeline.datatable, DataTable)
    assert pipeline.datatable.to_rows() == [{"value": 12}]
    assert pipeline.config.factor == 4
    assert pipeline.config.label == "demo"
    assert pipeline.config.nested["enabled"] is True
    assert pipeline.config.get("factor") == 4
    config_path.unlink(missing_ok=True)


def test_pipeline_restores_snapshot_when_stage_fails():
    import pandas as pd

    class FailingPipeline(Pipeline):
        stages = ["prepare", "explode"]

        def load_data(self):
            return pd.DataFrame({"value": [1, 2]})

        def prepare(self):
            self.datatable = self.datatable.assign(prepared=True)
            return self.datatable

        def explode(self):
            self.datatable = self.datatable.assign(bad=True)
            raise RuntimeError("boom")

    pipeline = FailingPipeline(name="failing")
    result = pipeline.run()

    assert result.failed
    assert list(pipeline.datatable.columns) == ["value", "prepared"]
    assert "bad" not in pipeline.datatable.columns


def test_pipeline_executes_registered_rollback_actions_on_failure():
    class SideEffectPipeline(Pipeline):
        stages = ["prepare", "explode"]

        def load_data(self):
            return 1

        def prepare(self):
            self.model.events = []
            self.model.events.append("prepare")
            self.on_rollback(lambda: self.model.events.append("undo"))
            self.datatable = 2
            return self.datatable

        def explode(self):
            self.model.events.append("explode")
            raise RuntimeError("boom")

    pipeline = SideEffectPipeline(name="side-effect")
    result = pipeline.run()

    assert result.failed
    assert pipeline.model.events == ["prepare", "explode", "undo"]




from __future__ import annotations

import importlib.util
import sys
from pathlib import Path


def _load_cli_module():
    spec = importlib.util.spec_from_file_location("soigia_cli", Path("soigia/cli.py"))
    module = importlib.util.module_from_spec(spec)
    assert spec.loader is not None
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    return module


def test_resolve_layout_target_accepts_preset_and_numeric():
    cli = _load_cli_module()

    pane_count, preset = cli._resolve_layout_target("4")
    assert pane_count == 4
    assert preset is None

    pane_count, preset = cli._resolve_layout_target("dev")
    assert pane_count == 3
    assert preset is not None
    assert preset.name == "dev"

    pane_count, preset = cli._resolve_layout_target("api")
    assert pane_count == 3
    assert preset is not None
    assert preset.name == "api"


def test_completion_script_mentions_powershell_completer():
    cli = _load_cli_module()

    script = cli._completion_script("powershell")

    assert "Register-ArgumentCompleter" in script
    assert "soigia" in script
    assert "suggest" in script
    assert "completion" in script
    assert "install-completion" in script
    assert "menu" in script


def test_completion_script_supports_bash_and_zsh():
    cli = _load_cli_module()

    bash_script = cli._completion_script("bash")
    zsh_script = cli._completion_script("zsh")

    assert "compgen" in bash_script
    assert "complete -F" in bash_script
    assert "#compdef" in zsh_script
    assert "_arguments" in zsh_script


def test_detect_shell_prefers_environment(monkeypatch):
    cli = _load_cli_module()
    monkeypatch.setenv("SHELL", "/usr/bin/zsh")
    assert cli._detect_shell() == "zsh"
    monkeypatch.setenv("SHELL", "/bin/bash")
    assert cli._detect_shell() == "bash"


def test_build_wt_command_uses_titles_and_segments(monkeypatch):
    cli = _load_cli_module()
    monkeypatch.setattr(cli, "_find_wt_executable", lambda: "wt.exe")

    pane_specs = cli._build_pane_specs(4, pane_titles=("dev", "tests", "logs", "notes"))
    command = cli._build_wt_command(pane_specs, r"C:\work")

    assert isinstance(command, list)
    assert command[0] == "wt.exe"
    assert "new-tab" in command
    assert "split-pane" in command
    assert "--title" in command
    assert "dev" in command
    assert "notes" in command
    assert ";" in command


def test_build_pane_specs_merges_custom_commands():
    cli = _load_cli_module()

    specs = cli._build_pane_specs(
        3,
        preset=cli.PRESETS["api"],
        pane_titles=("api", "tests", "docs"),
        pane_commands=[("powershell.exe", "-NoExit", "-Command", "Write-Host 'custom'")],
    )

    assert specs[0].title == "api"
    assert specs[0].command_template[0] == "powershell.exe"
    assert specs[1].title == "tests"
    assert specs[1].command_template
    assert specs[2].title == "docs"


def test_parse_pane_command_splits_command_line():
    cli = _load_cli_module()

    command = cli._parse_pane_command("powershell.exe -NoExit -Command \"Write-Host hi\"")

    assert command[0] == "powershell.exe"
    assert "-NoExit" in command
    assert "Write-Host hi" in command[-1]


def test_menu_routes_to_known_preset(monkeypatch):
    cli = _load_cli_module()
    responses = iter(["api", "launch"])
    monkeypatch.setattr(cli, "_menu_prompt", lambda title, options: next(responses))
    monkeypatch.setattr(cli, "_run_command", lambda command: None)

    result = cli._run_menu(object(), type("Args", (), {"cwd": None, "completion_shell": None, "profile": None})())

    assert result == 0


def test_preset_submenu_preview_and_dry_run(monkeypatch):
    cli = _load_cli_module()
    preset = cli.PRESETS["web"]
    args = type("Args", (), {"cwd": None})()

    monkeypatch.setattr(cli, "_menu_prompt", lambda title, options: "preview")
    monkeypatch.setattr(cli, "_preview_layout", lambda *args, **kwargs: None)
    assert cli._preset_submenu(object(), args, preset) == 0

    monkeypatch.setattr(cli, "_menu_prompt", lambda title, options: "dry-run")
    monkeypatch.setattr(cli, "_show_command_preview", lambda command: None)
    monkeypatch.setattr(cli, "_build_wt_command", lambda pane_specs, cwd: ["wt.exe", "new-tab"])
    assert cli._preset_submenu(object(), args, preset) == 0


def test_recent_layout_storage_roundtrip(monkeypatch):
    cli = _load_cli_module()
    recent_file = Path.cwd() / ".test_artifacts" / "recent.json"
    recent_file.parent.mkdir(parents=True, exist_ok=True)
    if recent_file.exists():
        recent_file.unlink()
    monkeypatch.setenv(cli.RECENT_ENV_PATH, str(recent_file))

    cli._remember_layout("api")
    cli._remember_layout("web")
    cli._remember_layout("api")

    assert cli._load_recent_layouts() == ["api", "web"]
    assert recent_file.exists()
    recent_file.unlink(missing_ok=True)


def test_menu_routes_to_recent(monkeypatch):
    cli = _load_cli_module()
    recent_file = Path.cwd() / ".test_artifacts" / "recent.json"
    recent_file.parent.mkdir(parents=True, exist_ok=True)
    recent_file.write_text('["api"]', encoding="utf-8")
    monkeypatch.setenv(cli.RECENT_ENV_PATH, str(recent_file))
    responses = iter(["recent", "1", "launch"])
    monkeypatch.setattr(cli, "_menu_prompt", lambda title, options: next(responses))
    monkeypatch.setattr(cli, "_run_command", lambda command: None)

    result = cli._run_menu(object(), type("Args", (), {"cwd": None, "completion_shell": None, "profile": None})())

    assert result == 0
    recent_file.unlink(missing_ok=True)


def test_install_completion_writes_marker_block():
    cli = _load_cli_module()
    profile = Path.cwd() / ".test_artifacts" / "cli-profile.ps1"
    profile.parent.mkdir(parents=True, exist_ok=True)
    if profile.exists():
        profile.unlink()

    installed_path = cli._install_completion("powershell", str(profile))

    text = installed_path.read_text(encoding="utf-8")
    assert cli.INSTALL_MARKER_START in text
    assert cli.INSTALL_MARKER_END in text
    assert "Register-ArgumentCompleter" in text
    installed_path.unlink(missing_ok=True)



from pathlib import Path

from soigia.config import ConfigStore, default_config, load_config, resolve_config_path, save_config, upsert_account


def test_config_store_roundtrip_and_account_helpers():
    artifact_dir = Path.cwd() / ".config_test_artifacts"
    artifact_dir.mkdir(exist_ok=True)
    path = artifact_dir / "config.yaml"
    try:
        config = default_config()
        upsert_account(
            config,
            "binance",
            "main",
            {
                "label": "Binance main",
                "enabled": True,
                "credentials": {"api_key": "key-123", "api_secret": "secret-456"},
                "settings": {"account_type": "spot"},
                "notes": "Primary exchange account",
            },
        )
        saved_path = save_config(config, path)
        loaded = load_config(saved_path)
        store = ConfigStore(saved_path)

        assert resolve_config_path(saved_path) == saved_path.resolve()
        assert store.get_account("binance", "main")["credentials"]["api_key"] == "key-123"
        assert loaded["accounts"]["binance"]["main"]["settings"]["account_type"] == "spot"
        assert store.get_accounts("binance")
    finally:
        if path.exists():
            path.unlink()



"""Tests for fake-data helpers."""

from soigia.datatable import DataTable


def test_from_faker_generates_fake_rows():
    table = DataTable.from_faker(
        {
            "name": "name",
            "email": "email",
            "age": lambda fake: fake.random_int(min=18, max=30),
            "city": "city",
        },
        count=5,
        seed=7,
    )

    assert isinstance(table, DataTable)
    assert len(table) == 5
    assert set(["name", "email", "age", "city"]).issubset(table.columns)
    assert table["age"].between(18, 30).all()


def test_from_factory_generates_fake_rows_from_batch_factory():
    class FakeUserFactory:
        @classmethod
        def build_batch(cls, size=1, **kwargs):
            rows = []
            for idx in range(size):
                rows.append(
                    {
                        "name": f"user_{idx}",
                        "age": 20 + idx,
                        "city": kwargs.get("city", "hanoi"),
                    }
                )
            return rows

    table = DataTable.from_factory(FakeUserFactory, count=3, city="saigon")

    assert isinstance(table, DataTable)
    assert table.to_rows() == [
        {"name": "user_0", "age": 20, "city": "saigon"},
        {"name": "user_1", "age": 21, "city": "saigon"},
        {"name": "user_2", "age": 22, "city": "saigon"},
    ]



"""Tests for the release automation helpers."""

import importlib.util
import sys
from pathlib import Path


def _load_release_module():
    spec = importlib.util.spec_from_file_location("release_script", Path("scripts/release.py"))
    module = importlib.util.module_from_spec(spec)
    assert spec.loader is not None
    spec.loader.exec_module(module)
    return module


def test_prompt_credentials_accepts_token_as_username(monkeypatch):
    release = _load_release_module()
    monkeypatch.delenv("TWINE_USERNAME", raising=False)
    monkeypatch.delenv("TWINE_PASSWORD", raising=False)
    monkeypatch.delenv("soigia-pip-user", raising=False)
    monkeypatch.delenv("soigia-pip-password", raising=False)
    monkeypatch.setattr("builtins.input", lambda prompt: "pypi-abc123")
    called = {}

    def fail_getpass(prompt):
        called["getpass"] = True
        return "should-not-run"

    monkeypatch.setattr(release.getpass, "getpass", fail_getpass)

    user, password = release._prompt_credentials("PyPI")

    assert user == "__token__"
    assert password == "pypi-abc123"
    assert "getpass" not in called


def test_prompt_credentials_accepts_token_in_environment(monkeypatch):
    release = _load_release_module()
    monkeypatch.setattr("builtins.input", lambda prompt: "someone")
    monkeypatch.setattr(release.getpass, "getpass", lambda prompt: "pypi-envtoken")
    user, password = release._prompt_credentials("PyPI")

    assert user == "__token__"
    assert password == "pypi-envtoken"


def test_build_upload_command_uses_repository_alias_by_default():
    release = _load_release_module()

    command = release._build_upload_command(["dist/a.whl"])

    assert command[:5] == [sys.executable, "-m", "twine", "upload", "--non-interactive"]
    assert "-r" in command
    assert "soigia" in command
    assert "--repository-url" not in command


def test_build_upload_command_uses_repository_url_when_provided():
    release = _load_release_module()

    command = release._build_upload_command(
        ["dist/a.whl"],
        repository="soigia",
        repository_url="https://test.pypi.org/legacy/",
        username="__token__",
        password="pypi-abc",
    )

    assert "--repository-url" in command
    assert "https://test.pypi.org/legacy/" in command
    assert "-r" not in command
    assert "-u" in command
    assert "-p" in command



"""Tests for join, schema validation, and auto-version helpers."""

from soigia.datatable import DataTable


def test_join_helpers_work():
    left = DataTable({"id": [1, 2, 3], "name": ["alice", "bob", "carol"]})
    right = DataTable({"id": [1, 2, 4], "city": ["hn", "sg", "dn"]})

    inner = left.inner_join(right, on="id")
    left_join = left.left_join(right, on="id")
    anti = left.anti_join(right, on="id")
    qs_join = left.objects.inner_join(right, on="id").to_df()

    assert list(inner["id"]) == [1, 2]
    assert list(inner["city"]) == ["hn", "sg"]
    assert list(left_join["id"]) == [1, 2, 3]
    assert list(anti["id"]) == [3]
    assert list(qs_join["name"]) == ["alice", "bob"]


def test_schema_helpers_work():
    table = DataTable({"name": ["alice"], "age": ["19"], "email": ["alice@example.com"]})

    ensured = table.ensure_columns("city", "country", fill_value="VN")
    coerced = table.coerce_types({"age": "int64"})
    validated = coerced.validate_schema(
        {
            "name": {"type": str, "nullable": False},
            "age": {"type": int, "min": 18, "max": 65, "nullable": False},
            "email": {"regex": r"@"},
            "city": {"required": False},
        }
    )

    assert list(ensured.columns) == ["name", "age", "email", "city", "country"]
    assert ensured["city"].iloc[0] == "VN"
    assert coerced["age"].dtype.kind in {"i", "u"}
    assert validated is coerced


def test_schema_validation_fails_for_bad_data():
    table = DataTable({"name": ["alice"], "age": [17]})

    try:
        table.validate_schema({"age": {"type": int, "min": 18}})
    except ValueError as exc:
        assert "minimum" in str(exc)
    else:  # pragma: no cover - defensive
        raise AssertionError("Expected schema validation to fail")


def test_auto_version_tracks_helper_transforms():
    table = DataTable({"name": ["alice"], "age": [19]})
    table.init_versions("initial")
    table.auto_version("auto")

    updated = table.fill_missing(0)
    ensured = updated.ensure_columns("city", fill_value="vn")

    assert table.is_auto_versioning() is True
    assert updated.version_count() == 2
    assert ensured.version_count() == 3
    assert updated.current_version()["label"] == "auto"
    assert ensured.current_version()["label"] == "auto"
    assert isinstance(ensured, DataTable)


def test_objects_delegate_new_helpers():
    table = DataTable({"id": [1, 2], "name": ["alice", "bob"], "age": ["19", "24"]})

    qs = table.objects.ensure_columns("city", fill_value="hn").coerce_types({"age": "int64"})
    materialized = qs.to_df()

    assert list(materialized["city"]) == ["hn", "hn"]
    assert materialized["age"].dtype.kind in {"i", "u"}



"""Integration coverage for DataTable end-to-end workflows."""

from pathlib import Path

import soigia.datatable as dtpkg
from soigia.datatable import DataTable


def test_end_to_end_dataframe_pipeline(monkeypatch):
    def fake_query(self, sql, backend, connection_string=None, **kwargs):
        return DataTable({"backend": [backend], "sql": [sql]})

    monkeypatch.setattr(dtpkg.DatabaseConnector, "query", fake_query)
    monkeypatch.setattr(dtpkg, "_DEFAULT_CONNECTOR", None)

    assert dtpkg.from_sqlite_db("select 1").iloc[0]["backend"] == "sqlite"
    assert dtpkg.from_sql("select 1", backend="sqlite").iloc[0]["backend"] == "sqlite"

    class DemoFactory:
        @classmethod
        def build_batch(cls, size=1, **kwargs):
            return [{"name": f"user_{i}", "age": 20 + i} for i in range(size)]

    fake_users = dtpkg.from_factory(DemoFactory, count=2)
    assert fake_users.shape == (2, 2)

    generated = dtpkg.from_faker({"name": "name", "city": "city"}, count=2, seed=3)
    assert generated.shape == (2, 2)

    base = DataTable.from_records(
        [
            {"id": 1, "name": "alice", "age": "19"},
            {"id": 2, "name": "bob", "age": "24"},
        ]
    ).auto_version("seed")

    prepared = base.clean_columns().ensure_columns("city", fill_value="unknown").to_numeric(["age"])
    assert prepared.validate_schema({"id": {"type": int}, "name": {"nullable": False}, "age": {"type": int}}).equals(
        prepared
    )

    enriched = prepared.join(DataTable({"id": [1, 2], "remote_city": ["hn", "sg"]}), on="id")
    assert "remote_city" in enriched.columns

    report = prepared.diff_report(prepared.copy(), key_columns=["id"])
    assert report["same_count"] == 2
    assert report["added_count"] == 0
    assert report["removed_count"] == 0
    assert report["has_changes"] is False

    manager = enriched.objects
    assert manager.values("id", "name").first()["id"] == 1
    assert manager.filter(age__gte=20).count() == 1
    assert manager.aggregate(total=("age", "sum")) == {"total": 43}
    assert manager.group_by("remote_city").annotate(total=("age", "sum")).to_df().shape[0] == 2
    assert manager.diff_report(joined := enriched.copy(), key_columns=["id"])["same_count"] == 2

    rendered = enriched.render_table("markdown")
    assert rendered.startswith("|")

    excel_path = Path("tests") / "_datatable_pipeline.xlsx"
    try:
        enriched.to_csv_table()
        assert enriched.to_markdown_table().startswith("|")
        assert enriched.to_html_table().startswith("<table")
    finally:
        if excel_path.exists():
            excel_path.unlink()



"""System, behavior, and user-acceptance style tests for DataTable."""

from pathlib import Path
from datetime import date
from types import SimpleNamespace

import pytest

import soigia.datatable as dtpkg
from soigia.datatable import DataTable


def test_system_pipeline_and_lookup_coverage(monkeypatch):
    base = DataTable(
        {
            "id": [1, 2, 3],
            "name": ["alice", "bob", "carol"],
            "score": [10, 20, 30],
            "when": ["2026-01-01", "2026-02-01", "2026-03-01"],
            "nullable": [1, None, 3],
        }
    )

    assert base.qs.count() == 3
    assert base.qs.values("name").first() == {"name": "alice"}
    assert base.qs.only().count() == 3
    assert base.qs.exclude_fields().count() == 3
    assert base.qs.filter(score__gt=10).count() == 2
    assert base.qs.filter(score__ne=20).count() == 2
    assert base.qs.filter(score__gte=20).count() == 2
    assert base.qs.filter(score__lt=30).count() == 2
    assert base.qs.filter(score__lte=20).count() == 2
    assert base.qs.filter(name__in=["alice", "carol"]).count() == 2
    assert base.qs.filter(name__contains="li").count() == 1
    assert base.qs.filter(name__icontains="ALI").count() == 1
    assert base.qs.filter(name__startswith="a").count() == 1
    assert base.qs.filter(name__istartswith="A").count() == 1
    assert base.qs.filter(name__endswith="l").count() == 1
    assert base.qs.filter(name__iendswith="L").count() == 1
    assert base.qs.filter(nullable__isnull=True).count() == 1
    assert base.qs.filter(score__range=(10, 20)).count() == 2
    assert base.qs.filter(score__between=(10, 20)).count() == 2
    assert base.qs.filter(when__year=2026).count() == 3
    assert base.qs.filter(when__month=2).count() == 1
    assert base.qs.filter(when__day=1).count() == 3
    assert base.qs.filter(when__date=date(2026, 1, 1)).count() == 1

    with pytest.raises(ValueError):
        base.qs.filter(score__unsupported=1)

    with pytest.raises(ValueError):
        base.qs.values_list("name", "score", flat=True)

    # Exercise the false branch in helper-specific render logic.
    monkeypatch.setattr(dtpkg, "openpyxl", None, raising=False)
    monkeypatch.setattr(dtpkg, "load_workbook", None, raising=False)
    monkeypatch.setattr(dtpkg, "Path", Path, raising=False)
    assert base.to_markdown_table().startswith("|")


def test_behavior_diff_and_schema_variants(monkeypatch):
    left = DataTable(
        {
            "id": [1, 2],
            "name": ["alice", "bob"],
            "age": [19, 24],
            "note": ["A", "B"],
        }
    ).auto_version("seed")
    right = DataTable(
        {
            "id": [1, 2, 3],
            "name": ["alice", "bobby", "carol"],
            "age": [19, 25, 31],
            "note": ["A", "C", "D"],
        }
    )

    assert left.diff_columns_against(right)["added_columns"] == []
    assert left.diff_added_rows(right, key_columns=["id"]).shape[0] == 1
    assert left.diff_removed_rows(right, key_columns=["id"]).shape[0] == 0
    assert left.diff_unchanged_rows(right, key_columns=["id"]).shape[0] == 1
    assert left.diff_changed_rows(right, key_columns=["id"]).shape[0] == 1
    assert left.diff_changed_cells(right, key_columns=["id"]).shape[0] == 3
    assert left.diff_changed_cells(right).shape[0] >= 0
    assert left.diff_report(right, key_columns=["id"])["has_changes"] is True
    assert left.different_rows(right, key_columns=["id"])["changed_count"] == 1
    assert left.compare_frames(right, key_columns=["id"])["changed_count"] == 1
    assert left.diff_versions()["row_delta"] == 0
    assert left.diff_records()["row_delta"] == 0
    assert left.diff_summary(key_columns=["id"])["changed_rows"] >= 0

    assert left.validate_schema({"name": {"nullable": False}}).equals(left)
    assert left.validate_schema({"id": int}).equals(left)
    assert left.validate_schema({"id": "ignored"}).equals(left)
    assert left.validate_schema({"missing": {"required": False}}).equals(left)
    assert left.validate_schema({"id": {"type": (int, float)}}).equals(left)
    assert left.validate_schema({"name": {"choices": ["alice", "bob"]}}).equals(left)
    assert left.validate_schema({"name": {"regex": "^[ab]"}}).equals(left)
    assert DataTable({"name": ["abcde"]}).validate_schema({"name": {"length": 5}}).equals(DataTable({"name": ["abcde"]}))
    assert left.validate_schema({"name": {"length": {"min": 3, "max": 5}}}).equals(left)
    assert left.validate_schema({"name": {"length": {"min": 2}}}).equals(left)
    assert left.validate_schema({"name": {"length": {"max": 10}}}).equals(left)
    assert left.validate_schema({"name": {"length": "ignored"}}).equals(left)

    with pytest.raises(ValueError):
        left.validate_schema({"name": {"nullable": False}}, strict=True)
    with pytest.raises(KeyError):
        left.validate_schema({"missing": {"required": True}})
    with pytest.raises(TypeError):
        left.validate_schema({"id": {"type": str}})
    with pytest.raises(ValueError):
        left.validate_schema({"name": {"choices": ["zzz"]}})
    with pytest.raises(ValueError):
        left.validate_schema({"name": {"regex": "^z"}})
    with pytest.raises(ValueError):
        left.validate_schema({"name": {"length": 2}})
    with pytest.raises(ValueError):
        left.validate_schema({"name": {"length": {"min": 6}}})
    with pytest.raises(ValueError):
        left.validate_schema({"name": {"length": {"max": 3}}})
    with pytest.raises(ValueError):
        left.validate_schema({"id": {"min": 20}})
    with pytest.raises(ValueError):
        left.validate_schema({"id": {"max": 1}})

    assert left.ensure_columns("extra").columns.tolist()[-1] == "extra"
    assert left.ensure_columns().equals(left)
    assert left.drop_empty_columns().equals(left)
    assert left.fill_missing(0).equals(left)
    assert left.fill_missing(0, columns=["name"]).equals(left)
    assert left.only("id").columns.tolist() == ["id"]
    assert left.only().equals(left)
    assert left.select("id").columns.tolist() == ["id"]
    assert left.select().equals(left)
    assert left.exclude_fields("note").columns.tolist() == ["id", "name", "age"]
    assert left.exclude_fields().equals(left)
    assert left.diff_changed_cells(right, key_columns=["id"]).shape[0] == 3
    assert left.diff_changed_cells(right).shape[0] >= 0
    with pytest.raises(KeyError):
        left.diff_changed_cells(right, key_columns=["missing"])
    with pytest.raises(KeyError):
        left.diff_report(right, key_columns=["missing"])
    assert left.objects.get_or_create(name="new", defaults={"age": 18})[1] is True
    assert left.objects.update_or_create(name="alice", defaults={"age": 20})[1] is False
    assert left.objects.group_by("id").aggregate(total="age")["total"]


def test_uat_writer_and_export_paths(monkeypatch):
    table = DataTable({"id": [1], "name": ["alice"], "age": [19]})

    assert table.objects.bulk_create([]).equals(table)
    updated = table.objects.bulk_update({"age": 20}, age__gte=19)
    assert updated["age"].iloc[0] == 20
    updated2, count = table.objects.bulk_update({"age": 20}, return_count=True, age__gte=19)
    assert count == 1
    assert updated2["age"].iloc[0] == 20
    assert table.objects.bulk_update({"note": "x"})["note"].iloc[0] == "x"
    assert table.objects.get_or_create(name="alice", defaults={"age": 19})[1] is False
    assert table.objects.update_or_create(name="alice", defaults={"age": 21})[1] is False
    assert table.objects.update_or_create(name="new", defaults={"age": 18})[1] is True

    assert table.objects.diff_report(table, key_columns=["id"])["has_changes"] is False
    assert table.objects.diff_columns_against(table)["column_order_changed"] is False
    assert table.objects.diff_changed_cells(table).shape[0] == 0

    fake_openpyxl = type("FakeOpenpyxlModule", (), {})()
    monkeypatch.setitem(dtpkg.__dict__, "openpyxl", fake_openpyxl)
    monkeypatch.setattr(DataTable, "to_excel", lambda self, output, **kwargs: Path(output).write_text("x", encoding="utf-8"), raising=False)
    path = Path("tests") / "_uat.xlsx"
    try:
        assert table.to_excel_table(path).exists()
    finally:
        if path.exists():
            path.unlink()


def test_render_and_import_fallback_branches(monkeypatch):
    table = DataTable({"a": [1], "b": [None]})

    monkeypatch.delitem(dtpkg.__dict__, "openpyxl", raising=False)
    monkeypatch.delitem(dtpkg.__dict__, "load_workbook", raising=False)
    assert table.drop_empty_columns().columns.tolist() == ["a"]
    assert table.render_table("rich") is not None
    with pytest.raises(ValueError):
        table.render_table("excel")
    with pytest.raises(ValueError):
        table.render_table("unknown")

    # direct helper coverage for unsupported lookup and null-equality branch
    assert dtpkg.DataTable._cell_equal(None, None) is True
    assert dtpkg.DataTable._cell_equal(float("nan"), float("nan")) is True
    with pytest.raises(ValueError):
        dtpkg._ORMQuerySet._apply_lookup(table["a"], "unsupported", 1)
    assert bool(dtpkg._ORMQuerySet._apply_lookup(table["a"], "exact", 1).iloc[0]) is True
    assert dtpkg.DataTable._cell_equal(dtpkg.pd.NA, dtpkg.pd.NA) is True


def test_queryset_render_delegation_branches(monkeypatch):
    table = DataTable({"name": ["alice"], "age": [19]})
    qs = table.qs
    path = Path("tests") / "_qs.xlsx"

    monkeypatch.setattr(DataTable, "to_rich_table", lambda self, *args, **kwargs: "RICH", raising=False)
    monkeypatch.setattr(DataTable, "to_excel_table", lambda self, *args, **kwargs: path, raising=False)

    assert qs.to_rich_table() == "RICH"
    assert qs.to_excel_table(path) == path


def test_google_client_builder_branches(monkeypatch):
    fake_gspread = SimpleNamespace(
        authorize=lambda credentials: {"client": credentials},
        service_account=lambda filename=None: {"service_account": filename or "default"},
    )

    class FakeCredentials:
        @classmethod
        def from_service_account_info(cls, info, scopes=None):
            return {"info": info, "scopes": scopes}

        @classmethod
        def from_service_account_file(cls, filename, scopes=None):
            return {"file": filename, "scopes": scopes}

    fake_google = SimpleNamespace(oauth2=SimpleNamespace(service_account=SimpleNamespace(Credentials=FakeCredentials)))
    monkeypatch.setitem(__import__("sys").modules, "gspread", fake_gspread)
    monkeypatch.setitem(__import__("sys").modules, "google", fake_google)
    monkeypatch.setitem(__import__("sys").modules, "google.oauth2", fake_google.oauth2)
    monkeypatch.setitem(
        __import__("sys").modules, "google.oauth2.service_account", fake_google.oauth2.service_account
    )

    client_from_json = dtpkg._build_google_client(credentials_json='{"x": 1}', scopes=["scope-1"])
    assert client_from_json["client"]["info"] == {"x": 1}
    client_from_file = dtpkg._build_google_client(service_account_file="creds.json", scopes=["scope-2"])
    assert client_from_file["client"]["file"] == "creds.json"
    assert dtpkg._build_google_client(client={"prebuilt": True}) == {"prebuilt": True}
    assert dtpkg._build_google_client() == {"service_account": "default"}



from soigia.db import Env, MemoryBackend, Model, PandasBackend, default_registry, fields


def test_db_package_exports_module():
    import soigia

    assert hasattr(soigia, "db")
    assert soigia.db.Model is Model


def test_db_memory_backend_queryset_and_relations():
    class Customer(Model):
        _name = "test.db.customer"

        name = fields.String(required=True)
        vip = fields.Boolean(default=False)

    class Order(Model):
        _name = "test.db.order"

        customer = fields.ForeignKey("test.db.customer")
        total = fields.Float(default=0)
        lines = fields.OneToMany("test.db.line", "order")

    class Line(Model):
        _name = "test.db.line"

        order = fields.ForeignKey("test.db.order")
        amount = fields.Float(default=0)

    env = Env(default_registry, MemoryBackend())

    customer = env["test.db.customer"].create({"name": "Alice", "vip": True})
    order = env["test.db.order"].create({"customer": customer, "total": 12.5})
    env["test.db.line"].create({"order": order, "amount": 5})
    env["test.db.line"].create({"order": order, "amount": 7.5})

    assert customer.name == "Alice"
    assert customer.vip is True
    assert order.customer.name == "Alice"
    assert order.lines.ids == [1, 2]
    assert env["test.db.order"].search([("total", "gt", 10)]).first().id == order.id
    assert env["test.db.order"].filter(total__gte=10).first().id == order.id


def test_db_inherit_extension_merges_fields_and_methods():
    class SaleOrder(Model):
        _name = "test.db.sale.order"

        name = fields.String(required=True)
        amount = fields.Float(default=0)

    class SaleOrderExtension(Model):
        _inherit = "test.db.sale.order"

        state = fields.String(default="draft")

        def action_confirm(self):
            self.state = "confirmed"
            self.save()
            return self.state

    env = Env(default_registry, MemoryBackend())
    order = env["test.db.sale.order"].create({"name": "SO-1", "amount": 100})

    assert order.state == "draft"
    assert order.action_confirm() == "confirmed"
    assert env["test.db.sale.order"].browse(order.id).state == "confirmed"


def test_db_pandas_backend_search_and_order():
    class Product(Model):
        _name = "test.db.product"

        name = fields.String(required=True)
        price = fields.Float(default=0)

    env = Env(default_registry, PandasBackend())
    env["test.db.product"].create({"name": "B", "price": 20})
    env["test.db.product"].create({"name": "A", "price": 10})

    products = env["test.db.product"].objects.filter(price__gte=10).order_by("-price").all()

    assert products.ids == [1, 2]
    assert [product.name for product in products] == ["B", "A"]
